# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import time
import warnings

import numpy as np
from numpy_groupies import aggregate as accumarray

import pyttb as ttb

from .pyttb_utils import *


def cp_apr(
    tensor,
    rank,
    algorithm="mu",
    stoptol=1e-4,
    stoptime=1e6,
    maxiters=1000,
    init="random",
    maxinneriters=10,
    epsDivZero=1e-10,
    printitn=1,
    printinneritn=0,
    kappa=0.01,
    kappatol=1e-10,
    epsActive=1e-8,
    mu0=1e-5,
    precompinds=True,
    inexact=True,
    lbfgsMem=3,
):
    """
    Compute non-negative CP with alternating Poisson regression.

    Parameters
    ----------
    tensor: :class:`pyttb.tensor` or :class:`pyttb.sptensor`
    rank: int
        Rank of the decomposition
    algorithm: str
        in {'mu', 'pdnr, 'pqnr'}
    stoptol: float
        Tolerance on overall KKT violation
    stoptime: float
        Maximum number of seconds to run
    maxiters: int
        Maximum number of iterations
    init: str or :class:`pyttb.ktensor`
        Initial guess
    maxinneriters: int
        Maximum inner iterations per outer iteration
    epsDivZero: float
        Safeguard against divide by zero
    printitn: int
        Print every n outer iterations, 0 for none
    printinneritn: int
        Print every n inner iterations
    kappa: int
        MU ALGORITHM PARAMETER: Offset to fix complementary slackness
    kappatol:
        MU ALGORITHM PARAMETER: Tolerance on complementary slackness
    epsActive: float
        PDNR & PQNR ALGORITHM PARAMETER: Bertsekas tolerance for active set
    mu0: float
        PDNR ALGORITHM PARAMETER: Initial Damping Parameter
    precompinds: bool
        PDNR & PQNR ALGORITHM PARAMETER: Precompute sparse tensor indices
    inexact: bool
        PDNR ALGORITHM PARAMETER: Compute inexact Newton steps
    lbfgsMem: int
        PQNR ALGORITHM PARAMETER: Precompute sparse tensor indices

    Returns
    -------
    M: :class:`pyttb.ktensor`
        Resulting ktensor from CP APR
    Minit: :class:`pyttb.ktensor`
        Initial Guess
    output: dict
        Additional output #TODO document this more appropriately

    """
    # Extract the number of modes in tensor X
    N = tensor.ndims

    assert rank > 0, "Number of components requested must be positive"

    # Check that the data is non-negative.
    tmp = tensor < 0.0
    assert (
        tmp.nnz == 0
    ), "Data tensor must be nonnegative for Poisson-based factorization"

    # Set up an initial guess for the factor matrices.
    if isinstance(init, ttb.ktensor):
        # User provided an initial ktensor; validate it
        assert init.ndims == N, "Initial guess does not have the right number of modes"
        assert (
            init.ncomponents == rank
        ), "Initial guess does not have the right number of componenets"
        for n in range(N):
            if init.shape[n] != tensor.shape[n]:
                assert False, "Mode {} of the initial guess is the wrong size".format(n)
            if np.min(init.factor_matrices[n]) < 0.0:
                assert False, "Initial guess has negative element in mode {}".format(n)
        if np.min(init.weights) < 0:
            assert False, "Initial guess has a negative ktensor weight"

    elif init.lower() == "random":
        factor_matrices = []
        for n in range(N):
            factor_matrices.append(np.random.uniform(0, 1, (tensor.shape[n], rank)))
        init = ttb.ktensor.from_factor_matrices(factor_matrices)

    # Call solver based on the couce of algorithm parameter, passing all the other input parameters
    if algorithm.lower() == "mu":
        M, output = tt_cp_apr_mu(
            tensor,
            rank,
            init,
            stoptol,
            stoptime,
            maxiters,
            maxinneriters,
            epsDivZero,
            printitn,
            printinneritn,
            kappa,
            kappatol,
        )
        output["algorithm"] = "mu"
    elif algorithm.lower() == "pdnr":
        M, output = tt_cp_apr_pdnr(
            tensor,
            rank,
            init,
            stoptol,
            stoptime,
            maxiters,
            maxinneriters,
            epsDivZero,
            printitn,
            printinneritn,
            epsActive,
            mu0,
            precompinds,
            inexact,
        )
        output["algorithm"] = "pdnr"
    elif algorithm.lower() == "pqnr":
        M, output = tt_cp_apr_pqnr(
            tensor,
            rank,
            init,
            stoptol,
            stoptime,
            maxiters,
            maxinneriters,
            epsDivZero,
            printitn,
            printinneritn,
            epsActive,
            lbfgsMem,
            precompinds,
        )
        output["algorithm"] = "pqnr"
    else:
        assert False, "{} is not a supported cp_als algorithm".format(algorithm)

    return M, init, output


def tt_cp_apr_mu(
    tensor,
    rank,
    init,
    stoptol,
    stoptime,
    maxiters,
    maxinneriters,
    epsDivZero,
    printitn,
    printinneritn,
    kappa,
    kappatol,
):
    """
    Compute nonnegative CP with alternating Poisson regression.

    Parameters
    ----------
    tensor: :class:`pyttb.tensor` or :class:`pyttb.sptensor`
    rank: int
        Rank of the decomposition
    init: :class:`pyttb.ktensor`
        Initial guess
    stoptol: float
        Tolerance on overall KKT violation
    stoptime: float
        Maximum number of seconds to run
    maxiters: int
        Maximum number of iterations
    maxinneriters: int
        Maximum inner iterations per outer iteration
    epsDivZero: float
        Safeguard against divide by zero
    printitn: int
        Print every n outer iterations, 0 for none
    printinneritn: int
        Print every n inner iterations
    kappa: int
        MU ALGORITHM PARAMETER: Offset to fix complementary slackness
    kappatol:
        MU ALGORITHM PARAMETER: Tolerance on complementary slackness

    Returns
    -------

    Notes
    -----
    REFERENCE: E. C. Chi and T. G. Kolda. On Tensors, Sparsity, and
    Nonnegative Factorizations, arXiv:1112.2414 [math.NA], December 2011,
    URL: http://arxiv.org/abs/1112.2414. Submitted for publication.

    """
    N = tensor.ndims

    # TODO I vote no duplicate error checking, copy error checking from cp_apr for initial guess here if disagree

    # Initialize output arrays
    # fnEvals = np.zeros((maxiters,))
    kktViolations = -np.ones((maxiters,))
    # TODO we initialize nInnerIters of size max outer iters?
    nInnerIters = np.zeros((maxiters,))
    # nzeros = np.zeros((maxiters,))
    nViolations = np.zeros((maxiters,))
    nTimes = np.zeros((maxiters,))

    # Set up for iteration - initializing M and Phi.
    # TODO replace with copy
    M = ttb.ktensor.from_tensor_type(init)
    M.normalize(normtype=1)
    Phi = []  # np.zeros((N,))#cell(N,1)
    for n in range(N):
        # TODO prepopulation Phi instead of appen should be faster
        Phi.append(np.zeros(M[n].shape))
    kktModeViolations = np.zeros((N,))

    if printitn > 0:
        print("\nCP_APR:\n")

    # Start the wall clock timer.
    start = time.time()

    # PDN-R and PQN-R benefit from precomputing sparse indices of X for each mode subproblem.
    # However, MU execution time barely changes, so the precomputer option is not offered.

    # Main loop: Iterate until convergence
    for iter in range(maxiters):
        isConverged = True
        for n in range(N):
            # Make adjustments to entries of M[n] that are violating complementary slackness conditions.
            # TODO both these zeros were 1 in matlab
            if iter > 0:
                V = (Phi[n] > 0) & (M[n] < kappatol)
                if np.any(V):
                    nViolations[iter] += 1
                    M.factor_matrices[n][V > 0] += kappa

            # Shift the weight from lambda to mode n
            M.redistribute(mode=n)

            # Calculate product of all matrices but the n-th
            # Sparse case only calculates entries corresponding to nonzeros in X
            Pi = calculatePi(tensor, M, rank, n, N)

            # Do the multiplicative updates
            for i in range(maxinneriters):
                # Count the inner iterations
                nInnerIters[iter] += 1

                # Calculate matrix for multiplicative update
                Phi[n] = calculatePhi(tensor, M, rank, n, Pi, epsDivZero)

                # Check for convergence
                kktModeViolations[n] = np.max(
                    np.abs(vectorizeForMu(np.minimum(M.factor_matrices[n], 1 - Phi[n])))
                )
                if kktModeViolations[n] < stoptol:
                    break
                else:
                    isConverged = False

                # Do the multiplicative update
                # TODO cannot update M[n] in this way
                M.factor_matrices[n] *= Phi[n]

                # Print status
                if printinneritn != 0 and divmod(i, printinneritn)[1] == 0:
                    print(
                        "\t\tMode = {}, Inner Iter = {}, KKT violation = {}\n".format(
                            n, i, kktModeViolations[n]
                        )
                    )

            # Shift weight from mode n back to lambda
            M.normalize(normtype=1, mode=n)

        kktViolations[iter] = np.max(kktModeViolations)
        if divmod(iter, printitn)[1] == 0:
            print(
                "\tIter {}: Inner Its = {} KKT violation = {}, nViolations = {}".format(
                    iter, nInnerIters[iter], kktViolations[iter], nViolations[iter]
                )
            )

        nTimes[iter] = time.time() - start

        # Check for convergence
        if isConverged:
            if printitn > 0:
                print("Exiting because all subproblems reached KKT tol.\n")
            break
        if nTimes[iter] > stoptime:
            if printitn > 0:
                print("Exiting because time limit exceeded.\n")
            break

    t_stop = time.time() - start

    # Clean up final result
    M.normalize(sort=True, normtype=1)

    obj = tt_loglikelihood(tensor, M)

    if printitn > 0:
        normTensor = tensor.norm()
        normresidual = np.sqrt(
            normTensor**2 + M.norm() ** 2 - 2 * tensor.innerprod(M)
        )
        fit = 1 - (normresidual / normTensor)  # fraction explained by model
        print("===========================================\n")
        print(" Final log-likelihood = {} \n".format(obj))
        print(" Final least squares fit = {} \n".format(fit))
        print(" Final KKT violation = {}\n".format(kktViolations[iter]))
        print(" Total inner iterations = {}\n".format(sum(nInnerIters)))
        print(" Total execution time = {} secs\n".format(t_stop))

    output = {}
    output["params"] = (
        stoptol,
        stoptime,
        maxiters,
        maxinneriters,
        epsDivZero,
        printitn,
        printinneritn,
        kappa,
        kappatol,
    )
    output["kktViolations"] = kktViolations[: iter + 1]
    output["nInnerIters"] = nInnerIters[: iter + 1]
    output["nViolations"] = nViolations[: iter + 1]
    output["nTotalIters"] = np.sum(nInnerIters)
    output["times"] = nTimes[: iter + 1]
    output["totalTime"] = t_stop
    output["obj"] = obj

    return M, output


def tt_cp_apr_pdnr(
    tensor,
    rank,
    init,
    stoptol,
    stoptime,
    maxiters,
    maxinneriters,
    epsDivZero,
    printitn,
    printinneritn,
    epsActive,
    mu0,
    precompinds,
    inexact,
):
    """
    Compute nonnegative CP with alternating Poisson regression
    computes an estimate of the best rank-R
    CP model of a tensor X using an alternating Poisson regression.
    The algorithm solves "row subproblems" in each alternating subproblem,
    using a Hessian of size R^2.

    Parameters
    ----------
    # TODO it looks like this method of define union helps the typ hinting better than or
    tensor: Union[:class:`pyttb.tensor`,:class:`pyttb.sptensor`]
    rank: int
        Rank of the decomposition
    init: str or :class:`pyttb.ktensor`
        Initial guess
    stoptol: float
        Tolerance on overall KKT violation
    stoptime: float
        Maximum number of seconds to run
    maxiters: int
        Maximum number of iterations
    maxinneriters: int
        Maximum inner iterations per outer iteration
    epsDivZero: float
        Safeguard against divide by zero
    printitn: int
        Print every n outer iterations, 0 for none
    printinneritn: int
        Print every n inner iterations
    epsActive: float
        PDNR & PQNR ALGORITHM PARAMETER: Bertsekas tolerance for active set
    mu0: float
        PDNR ALGORITHM PARAMETER: Initial Damping Parameter
    precompinds: bool
        PDNR & PQNR ALGORITHM PARAMETER: Precompute sparse tensor indices
    inexact: bool
        PDNR ALGORITHM PARAMETER: Compute inexact Newton steps

    Returns
    -------
    # TODO detail return dictionary

    Notes
    -----
    REFERENCE: Samantha Hansen, Todd Plantenga, Tamara G. Kolda.
    Newton-Based Optimization for Nonnegative Tensor Factorizations,
    arXiv:1304.4964 [math.NA], April 2013,
    URL: http://arxiv.org/abs/1304.4964. Submitted for publication.

    """
    # Extract the number of modes in tensor X
    N = tensor.ndims

    # If the initial guess has any rows of all zero elements, then modify so the row subproblem is not taking log(0).
    # Values will be restored to zero later if the unfolded X for the row has no zeros.
    for n in range(N):
        rowsum = np.sum(init[n], axis=1)
        tmpIdx = np.where(rowsum == 0)[0]
        if tmpIdx.size != 0:
            init[n][tmpIdx, 0] = 1e-8

    # Start with the initial guess, normalized using the vector L1 norm
    # TODO replace with copy
    M = ttb.ktensor.from_tensor_type(init)
    M.normalize(normtype=1)

    # Sparse tensor flag affects how Pi and Phi are computed.
    if isinstance(tensor, ttb.sptensor):
        isSparse = True
    else:
        isSparse = False

    # Initialize output arrays
    fnEvals = np.zeros((maxiters, 1))
    fnVals = np.zeros((maxiters, 1))
    kktViolations = -np.ones((maxiters, 1))
    nInnerIters = np.zeros((maxiters, 1))
    nzeros = np.zeros((maxiters, 1))
    times = np.zeros((maxiters, 1))

    if printitn > 0:
        print("\nCP_PDNR (alternating Poisson regression using damped Newton)\n")

    dispLineWarn = printinneritn > 0

    # Start the wall clock timer.
    start = time.time()

    if isSparse and precompinds:
        # Precompute sparse index sets for all the row subproblems.
        # Takes more memory but can cut exectuion time significantly in some cases.
        if printitn > 0:
            print("\tPrecomuting sparse index sets...")
        sparseIx = []
        for n in range(N):
            num_rows = M[n].shape[0]
            row_indices = []
            for jj in range(num_rows):
                row_indices.append(np.where(tensor.subs[:, n] == jj)[0])
            sparseIx.append(row_indices)

        if printitn > 0:
            print("done\n")

    e_vec = np.ones((1, rank))

    rowsubprobStopTol = stoptol

    # Main loop: iterate until convergence or a max threshold is reached
    for iter in range(maxiters):
        isConverged = True
        kktModeViolations = np.zeros((N,))
        countInnerIters = np.zeros((N,))

        # Alternate thru each factor matrix, A_1, A_2, ..., A_N.
        for n in range(N):
            # Shift the weight from lambda to mode n.
            M.redistribute(mode=n)

            # calculate khatri-rao product of all matrices but the n-th
            if isSparse == False:
                # Data is not a sparse tensor.
                Pi = ttb.tt_calcpi_prowsubprob(tensor, M, rank, n, N, isSparse)
                X_mat = ttb.tt_to_dense_matrix(tensor, n)

            num_rows = M[n].shape[0]
            isRowNOTconverged = np.zeros((num_rows,))

            # Loop over the row subproblems in mode n.
            for jj in range(num_rows):
                # Initialize the damped Hessian parameter for the row subproblem.
                mu = mu0

                # Get data values for row jj of matricized mode n,
                if isSparse:
                    # Data is a sparse tensor
                    if not precompinds:
                        sparse_indices = np.where(tensor.subs[:, n] == jj)[0]
                    else:
                        sparse_indices = sparseIx[n][jj]

                    if sparse_indices.size == 0:
                        # The row jj of matricized tensor X in mode n is empty
                        M.factor_matrices[n][jj, :] = 0
                        continue

                    x_row = tensor.vals[sparse_indices]

                    # Calculate just the columns of Pi needed for this row.
                    Pi = ttb.tt_calcpi_prowsubprob(
                        tensor, M, rank, n, N, isSparse, sparse_indices
                    )

                else:
                    x_row = X_mat[jj, :]

                # Get current values of the row subproblem variables.
                m_row = M[n][jj, :]

                # Iteratively solve the row subproblem with projected Newton steps.
                if inexact and iter == 1:
                    innerIterMaximum = 2
                else:
                    innerIterMaximum = maxinneriters

                for i in range(innerIterMaximum):
                    # Calculate the gradient.
                    [phi_row, ups_row] = calc_partials(
                        isSparse, Pi, epsDivZero, x_row, m_row
                    )
                    gradM = (e_vec - phi_row).transpose()

                    # Compute the row subproblem kkt_violation.

                    # Note experiments in the original paper used:
                    # kkt_violation = np.norm(np.abs(np.minimum(m_row, gradM.transpose())))

                    # We now use \| KKT \|_{inf}:
                    kkt_violation = np.max(
                        np.abs(np.minimum(m_row, gradM.transpose()[0]))
                    )

                    # Report largest row subproblem initial violation
                    if i == 0 and kkt_violation > kktModeViolations[n]:
                        kktModeViolations[n] = kkt_violation

                    if printinneritn > 0 and np.mod(i, printinneritn) == 0:
                        print("\tMode = {}, Row = {}, InnerIt = {}".format(n, jj, i))

                        if i == 0:
                            print(", RowKKT = {}\n".format(kkt_violation))
                        else:
                            print(
                                ", RowKKT = {}, RowObj = {}\n".format(
                                    kkt_violation, -f_new
                                )
                            )

                    # Check for row subproblem convergence.
                    if kkt_violation < stoptol:
                        break
                    else:
                        # Not converged, so m_row will be modified.
                        isRowNOTconverged[jj] = 1

                    # Calculate the search direction
                    # TODO clean up reshaping gradM to row
                    search_dir, predicted_red = getSearchDirPdnr(
                        Pi, ups_row, rank, gradM.transpose()[0], m_row, mu, epsActive
                    )

                    # Perform a projected linesearch and update variables.
                    # Start from a unit step length, decrease by 1/2,
                    # stop with sufficicent decrease of 1.0e-4 or at most 10 steps.
                    (
                        m_rowNew,
                        f_old,
                        f_unit,
                        f_new,
                        num_evals,
                    ) = ttb.tt_linesearch_prowsubprob(
                        search_dir.transpose()[0],
                        gradM.transpose(),
                        m_row,
                        1,
                        1 / 2,
                        10,
                        1.0e-4,
                        isSparse,
                        x_row,
                        Pi,
                        phi_row,
                        dispLineWarn,
                    )
                    fnEvals[iter] += num_evals
                    m_row = m_rowNew

                    # Update damping parameter mu based on the unit step length, which is returned in f_unit
                    actual_red = f_old - f_unit
                    rho = actual_red / -predicted_red
                    if predicted_red == 0:
                        mu *= 10
                    elif rho < 1 / 4:
                        mu *= 7 / 2
                    elif rho > 3 / 4:
                        mu *= 2 / 7

                M.factor_matrices[n][jj, :] = m_row
                countInnerIters[n] += i

            # Test if all row subproblems have converged, which means that no varibales in this more were changed.
            if np.sum(isRowNOTconverged) != 0:
                isConverged = False

            # Shift weight from mode n back to lambda.
            M.normalize(mode=n, normtype=1)

            # Total number of inner iterations for a given outer iteration, totalled across all modes and all
            # row subproblems in each mode
            nInnerIters[iter] += countInnerIters[n]

        # Save output items for the outer iteration.
        num_zero = 0
        for n in range(N):
            num_zero += np.count_nonzero(M[n] == 0)  # [0].size

        nzeros[iter] = num_zero
        kktViolations[iter] = np.max(kktModeViolations)

        if inexact:
            rowsubprobStopTol = np.maximum(stoptol, kktViolations[iter]) / 100.0

            # Print outer iteration status.
            if printitn > 0 and np.mod(iter, printitn) == 0:
                fnVals[iter] = -tt_loglikelihood(tensor, M)
                print(
                    "{}. Ttl Inner Its: {}, KKT viol = {}, obj = {}, nz: {}\n".format(
                        iter,
                        nInnerIters[iter],
                        kktViolations[iter],
                        fnVals[iter],
                        num_zero,
                    )
                )

        times[iter] = time.time() - start

        # Check for convergence
        if isConverged and inexact == False:
            break
        if isConverged and inexact and rowsubprobStopTol <= stoptol:
            break
        if times[iter] > stoptime:
            print("EXiting because time limit exceeded\n")
            break

    t_stop = time.time() - start

    # Clean up final result
    M.normalize(sort=True, normtype=1)

    obj = tt_loglikelihood(tensor, M)

    if printitn > 0:
        normTensor = tensor.norm()
        normresidual = np.sqrt(
            normTensor**2 + M.norm() ** 2 - 2 * tensor.innerprod(M)
        )
        fit = 1 - (normresidual / normTensor)  # fraction explained by model
        print("===========================================\n")
        print(" Final log-likelihood = {} \n".format(obj))
        print(" Final least squares fit = {} \n".format(fit))
        print(" Final KKT violation = {}\n".format(kktViolations[iter]))
        print(" Total inner iterations = {}\n".format(sum(nInnerIters)))
        print(" Total execution time = {} secs\n".format(t_stop))

    output = {}
    output["params"] = (
        stoptol,
        stoptime,
        maxiters,
        maxinneriters,
        epsDivZero,
        printitn,
        printinneritn,
        epsActive,
        mu0,
        precompinds,
        inexact,
    )
    output["kktViolations"] = kktViolations[: iter + 1]
    output["obj"] = obj
    output["fnEvals"] = fnEvals[: iter + 1]
    output["fnVals"] = fnVals[: iter + 1]
    output["nInnerIters"] = nInnerIters[: iter + 1]
    output["nZeros"] = nzeros[: iter + 1]
    output["times"] = times[: iter + 1]
    output["totalTime"] = t_stop

    return M, output


def tt_cp_apr_pqnr(
    tensor,
    rank,
    init,
    stoptol,
    stoptime,
    maxiters,
    maxinneriters,
    epsDivZero,
    printitn,
    printinneritn,
    epsActive,
    lbfgsMem,
    precompinds,
):
    """
    Compute nonnegative CP with alternating Poisson regression.

    tt_cp_apr_pdnr computes an estimate of the best rank-R
    CP model of a tensor X using an alternating Poisson regression.
    The algorithm solves "row subproblems" in each alternating subproblem,
    using a Hessian of size R^2.
    The function is typically called by cp_apr.

    The model is solved by nonlinear optimization, and the code literally
    minimizes the negative of log-likelihood.  However, printouts to the
    console reverse the sign to show maximization of log-likelihood.


    mu0: float
        PDNR ALGORITHM PARAMETER: Initial Damping Parameter
    precompinds: bool
        PDNR & PQNR ALGORITHM PARAMETER: Precompute sparse tensor indices
    inexact: bool
        PDNR ALGORITHM PARAMETER: Compute inexact Newton steps

    Parameters
    ----------
    tensor: Union[:class:`pyttb.tensor`,:class:`pyttb.sptensor`]
    rank: int
        Rank of the decomposition
    init: str or :class:`pyttb.ktensor`
        Initial guess
    stoptol: float
        Tolerance on overall KKT violation
    stoptime: float
        Maximum number of seconds to run
    maxiters: int
        Maximum number of iterations
    maxinneriters: int
        Maximum inner iterations per outer iteration
    epsDivZero: float
        Safeguard against divide by zero
    printitn: int
        Print every n outer iterations, 0 for none
    printinneritn: int
        Print every n inner iterations
    epsActive: float
        PDNR & PQNR ALGORITHM PARAMETER: Bertsekas tolerance for active set
    lbfgsMem: int
        Number of vector pairs to store for L-BFGS
    precompinds

    Returns
    -------
    # TODO detail return dictionary

    Notes
    -----
    REFERENCE: Samantha Hansen, Todd Plantenga, Tamara G. Kolda.
    Newton-Based Optimization for Nonnegative Tensor Factorizations,
    arXiv:1304.4964 [math.NA], April 2013,
    URL: http://arxiv.org/abs/1304.4964. Submitted for publication.

    """
    # TODO first ~100 lines are identical to PDNR, consider abstracting just the algorithm portion
    # Extract the number of modes in data tensor
    N = tensor.ndims

    # If the initial guess has any rows of all zero elements, then modify so the row subproblem is not taking log(0).
    # Values will be restored to zero later if the unfolded X for the row has no zeros.
    for n in range(N):
        rowsum = np.sum(init[n], axis=1)
        tmpIdx = np.where(rowsum == 0)[0]
        if tmpIdx.size != 0:
            init[n][tmpIdx, 0] = 1e-8

    # Start with the initial guess, normalized using the vector L1 norm
    # TODO replace with copy
    M = ttb.ktensor.from_tensor_type(init)
    M.normalize(normtype=1)

    # Sparse tensor flag affects how Pi and Phi are computed.
    if isinstance(tensor, ttb.sptensor):
        isSparse = True
    else:
        isSparse = False

    # Initialize output arrays
    fnEvals = np.zeros((maxiters, 1))
    fnVals = np.zeros((maxiters, 1))
    kktViolations = -np.ones((maxiters, 1))
    nInnerIters = np.zeros((maxiters, 1))
    nzeros = np.zeros((maxiters, 1))
    times = np.zeros((maxiters, 1))

    if printitn > 0:
        print("\nCP_PQNR (alternating Poisson regression using quasi-Newton)\n")

    dispLineWarn = printinneritn > 0

    # Start the wall clock timer.
    start = time.time()

    if isSparse and precompinds:
        # Precompute sparse index sets for all the row subproblems.
        # Takes more memory but can cut exectuion time significantly in some cases.
        if printitn > 0:
            print("\tPrecomuting sparse index sets...")
        sparseIx = []
        for n in range(N):
            num_rows = M[n].shape[0]
            row_indices = []
            for jj in range(num_rows):
                row_indices.append(np.where(tensor.subs[:, n] == jj)[0])
            sparseIx.append(row_indices)

        if printitn > 0:
            print("done\n")

    # Main loop: iterate until convergence or a max threshold is reached
    for iter in range(maxiters):
        isConverged = True
        kktModeViolations = np.zeros((N,))
        countInnerIters = np.zeros((N,))

        # Alternate thru each factor matrix, A_1, A_2, ..., A_N.
        for n in range(N):
            # Shift the weight from lambda to mode n.
            M.redistribute(mode=n)

            # calculate khatri-rao product of all matrices but the n-th
            if isSparse == False:
                # Data is not a sparse tensor.
                Pi = ttb.tt_calcpi_prowsubprob(tensor, M, rank, n, N, isSparse)
                X_mat = ttb.tt_to_dense_matrix(tensor, n)

            num_rows = M[n].shape[0]
            isRowNOTconverged = np.zeros((num_rows,))

            # Loop over the row subproblems in mode n.
            for jj in range(num_rows):
                # Get data values for row jj of matricized mode n,
                if isSparse:
                    # Data is a sparse tensor
                    if not precompinds:
                        sparse_indices = np.where(tensor.subs[:, n] == jj)[0]
                    else:
                        sparse_indices = sparseIx[n][jj]

                    if sparse_indices.size == 0:
                        # The row jj of matricized tensor X in mode n is empty
                        M.factor_matrices[n][jj, :] = 0
                        continue

                    x_row = tensor.vals[sparse_indices]

                    # Calculate just the columns of Pi needed for this row.
                    Pi = ttb.tt_calcpi_prowsubprob(
                        tensor, M, rank, n, N, isSparse, sparse_indices
                    )

                else:
                    x_row = X_mat[jj, :]

                # Get current values of the row subproblem variables.
                m_row = M[n][jj, :]

                # Initialize L-BFGS storage for the row subproblem.
                delm = np.zeros((rank, lbfgsMem))
                delg = np.zeros((rank, lbfgsMem))
                rho = np.zeros((lbfgsMem,))
                lbfgsPos = 0
                m_rowOLD = []
                gradOLD = []

                # Iteratively solve the row subproblem with projected quasi-Newton steps
                for i in range(maxinneriters):
                    # Calculate the gradient.
                    gradM, phi_row = calc_grad(isSparse, Pi, epsDivZero, x_row, m_row)
                    if i == 0:
                        # Note from MATLAB tensortoolbox
                        # Original cp_aprPQN_row code (and plb_row) does a gradient
                        # step to prime the L-BFGS approximation.  However, it means
                        # a row subproblem that already converged wastes time
                        # doing a gradient step before checking KKT conditions.
                        # TODO: fix in a future release.
                        m_rowOLD = m_row
                        gradOLD = gradM
                        m_row, f, f_unit, f_new, num_evals = tt_linesearch_prowsubprob(
                            -gradM.transpose(),
                            gradM.transpose(),
                            m_rowOLD,
                            1,
                            1 / 2,
                            10,
                            1e-4,
                            isSparse,
                            x_row,
                            Pi,
                            phi_row,
                            dispLineWarn,
                        )
                        fnEvals[iter] += num_evals
                        gradM, phi_row = calc_grad(
                            isSparse, Pi, epsDivZero, x_row, m_row
                        )

                    # Compute the row subproblem kkt_violation.

                    # Note experiments in the original paper used:
                    # kkt_violation = np.norm(np.abs(np.minimum(m_row, gradM.transpose())))

                    # We now use \| KKT \|_{inf}:
                    kkt_violation = np.max(np.abs(np.minimum(m_row, gradM)))
                    # print("Intermediate Printing m_row: {}\n and gradM{}".format(m_row, gradM))

                    # Report largest row subproblem initial violation
                    if i == 0 and kkt_violation > kktModeViolations[n]:
                        kktModeViolations[n] = kkt_violation

                    if printinneritn > 0 and np.mod(i, printinneritn) == 0:
                        print("\tMode = {}, Row = {}, InnerIt = {}".format(n, jj, i))

                        if i == 0:
                            print(", RowKKT = {}\n".format(kkt_violation))
                        else:
                            print(
                                ", RowKKT = {}, RowObj = {}\n".format(
                                    kkt_violation, -f_new
                                )
                            )

                    # Check for row subproblem convergence.
                    if kkt_violation < stoptol:
                        break
                    else:
                        # Not converged, so m_row will be modified.
                        isRowNOTconverged[jj] = 1

                    # Update the L-BFGS approximation.
                    tmp_delm = m_row - m_rowOLD
                    tmp_delg = gradM - gradOLD
                    tmp_delm_dot = tmp_delm.dot(tmp_delg.transpose())
                    if not np.any(tmp_delm_dot == 0):
                        tmp_rho = 1 / tmp_delm_dot
                        delm[:, lbfgsPos] = tmp_delm
                        delg[:, lbfgsPos] = tmp_delg
                        rho[lbfgsPos] = tmp_rho
                    else:
                        # Rho is required to be postive; if not, then skip the L-BFGS update pair. The recommended
                        # safeguard for full BFGS is Powell damping, but not clear how to damp in 2-loop L-BFGS
                        if dispLineWarn:
                            warnings.warn(
                                "WARNING: skipping L-BFGS update, rho whould be 1 / {}".format(
                                    tmp_delm * tmp_delg
                                )
                            )
                        # Roll back lbfgsPos since it will increment later.
                        if lbfgsPos == 0:
                            if rho[lbfgsMem - 1] > 0:
                                lbfgsPos = lbfgsMem - 1
                            else:
                                # Fatal error, should not happen.
                                assert False, "ERROR: L-BFGS first iterate is bad"

                        else:
                            lbfgsPos -= 1

                    # Calculate search direction
                    search_dir = getSearchDirPqnr(
                        m_row,
                        gradM,
                        epsActive,
                        delm,
                        delg,
                        rho,
                        lbfgsPos,
                        i,
                        dispLineWarn,
                    )

                    lbfgsPos = np.mod(lbfgsPos, lbfgsMem)

                    m_rowOLD = m_row
                    gradOLD = gradM

                    # Perform a projected linesearch and update variables.
                    # Start from a unit step length, decrease by 1/2,
                    # stop with sufficicent decrease of 1.0e-4 or at most 10 steps.
                    m_row, f, f_unit, f_new, num_evals = ttb.tt_linesearch_prowsubprob(
                        search_dir.transpose()[0],
                        gradOLD.transpose(),
                        m_rowOLD,
                        1,
                        1 / 2,
                        10,
                        1.0e-4,
                        isSparse,
                        x_row,
                        Pi,
                        phi_row,
                        dispLineWarn,
                    )
                    fnEvals[iter] += num_evals

                M.factor_matrices[n][jj, :] = m_row
                countInnerIters[n] += i

            # Test if all row subproblems have converged, which means that no varibales in this more were changed.
            if np.sum(isRowNOTconverged) != 0:
                isConverged = False

            # Shift weight from mode n back to lambda.
            M.normalize(mode=n, normtype=1)

            # Total number of inner iterations for a given outer iteration, totalled across all modes and all
            # row subproblems in each mode
            nInnerIters[iter] += countInnerIters[n]

        # Save output items for the outer iteration.
        num_zero = 0
        for n in range(N):
            num_zero += np.count_nonzero(M[n] == 0)  # [0].size

        nzeros[iter] = num_zero
        kktViolations[iter] = np.max(kktModeViolations)

        # Print outer iteration status.
        if printitn > 0 and np.mod(iter, printitn) == 0:
            fnVals[iter] = -tt_loglikelihood(tensor, M)
            print(
                "{}. Ttl Inner Its: {}, KKT viol = {}, obj = {}, nz: {}\n".format(
                    iter, nInnerIters[iter], kktViolations[iter], fnVals[iter], num_zero
                )
            )

        times[iter] = time.time() - start

        # Check for convergence
        if isConverged:
            break
        if times[iter] > stoptime:
            print("Exiting because time limit exceeded\n")
            break

    t_stop = time.time() - start

    # Clean up final result
    M.normalize(sort=True, normtype=1)

    obj = tt_loglikelihood(tensor, M)

    if printitn > 0:
        normTensor = tensor.norm()
        normresidual = np.sqrt(
            normTensor**2 + M.norm() ** 2 - 2 * tensor.innerprod(M)
        )
        fit = 1 - (normresidual / normTensor)  # fraction explained by model
        print("===========================================\n")
        print(" Final log-likelihood = {} \n".format(obj))
        print(" Final least squares fit = {} \n".format(fit))
        print(" Final KKT violation = {}\n".format(kktViolations[iter]))
        print(" Total inner iterations = {}\n".format(sum(nInnerIters)))
        print(" Total execution time = {} secs\n".format(t_stop))

    output = {}
    output["params"] = (
        stoptol,
        stoptime,
        maxiters,
        maxinneriters,
        epsDivZero,
        printitn,
        printinneritn,
        epsActive,
        lbfgsMem,
        precompinds,
    )
    output["kktViolations"] = kktViolations[: iter + 1]
    output["obj"] = obj
    output["fnEvals"] = fnEvals[: iter + 1]
    output["fnVals"] = fnVals[: iter + 1]
    output["nInnerIters"] = nInnerIters[: iter + 1]
    output["nZeros"] = nzeros[: iter + 1]
    output["times"] = times[: iter + 1]
    output["totalTime"] = t_stop

    return M, output


# PDNR helper functions
def tt_calcpi_prowsubprob(
    Data, Model, rank, factorIndex, ndims, isSparse=False, sparse_indices=None
):
    """
    Compute Pi for a row subproblem.

    Parameters
    ----------
    Data :class:`pyttb.sptensor` or :class:`pyttb.tensor`
    isSparse: bool
    Model :class:`pyttb.ktensor`
    rank: int
    factorIndex: int
    ndims: int
    sparse_indices: list
        Indices of row subproblem nonzero elements

    Returns
    -------
    Pi: :class:`numpy.ndarray`

    See Also
    --------
    :class:`pyttb.calculatePi`

    """
    # TODO: this can probably be merged with general calculate pi, where default for sparse_indices is slice(None,None,None)
    if isSparse:
        # Data is a sparse tensor. Compute Pi for the row problem specified by sparse_indices
        num_row_nnz = len(sparse_indices)

        Pi = np.ones((num_row_nnz, rank))
        for i in np.setdiff1d(np.arange(ndims), factorIndex).astype(int):
            Pi *= Model[i][Data.subs[sparse_indices, i], :]
    else:
        Pi = ttb.khatrirao(
            Model.factor_matrices[:factorIndex]
            + Model.factor_matrices[factorIndex + 1 : ndims + 1],
            reverse=True,
        )

    return Pi


def calc_partials(isSparse, Pi, epsilon, data_row, model_row):
    """
    Compute derivative quantities for a PDNR row subproblem.

    Parameters
    ----------
    isSparse: bool
    Pi: :class:`numpy.ndarray`
    epsilon: float
        Prevent division by zero
    data_row: :class:`numpy.ndarray`
    model_row: :class:`numpy.ndarray`

    Returns
    -------
    phi_row: :class:`numpy.ndarray`
        gradient of row subproblem, except for a constant \n
        :math:`phi\_row[r] = \sum_{j=1}^{J_n}\\frac{x_j\pi_{rj}}{\sum_i^R b_i\pi_{ij}}`
    ups_row: :class:`numpy.ndarray`
        intermediate quantity (upsilon) used for second derivatives  \n
        :math:`ups\_row[j] = \\frac{x_j}{\\left(\sum_i^R b_i\pi_{ij}\\right)^2}`

    """
    if isSparse:
        data_row = data_row.transpose()[0]
    v = model_row.dot(Pi.transpose())
    w = data_row.transpose() / np.maximum(v, epsilon)
    phi_row = w.dot(Pi)
    u = v**2
    ups_row = data_row.transpose() / np.maximum(u, epsilon)
    return phi_row, ups_row


def getSearchDirPdnr(Pi, ups_row, rank, gradModel, model_row, mu, epsActSet):
    """
    Compute the search direction for PDNR using a two-metric projection with damped Hessian

    Parameters
    ----------
    Pi: :class:`numpy.ndarray`
    ups_row: :class:`numpy.ndarray`
        intermediate quantity (upsilon) used for second derivatives
    rank: int
        number of variables for the row subproblem
    gradModel:  :class:`numpy.ndarray`
        gradient vector for the row subproblem
    model_row: :class:`numpy.ndarray`
        vector of variables for the row subproblem
    mu: float
        damping parameter
    epsActSet: float
        Bertsekas tolerance for active set determination

    Returns
    -------
    search_dir: :class:`numpy.ndarray`
        search direction vector
    pred_red: :class:`numpy.ndarray`
        predicted reduction in quadratic model
    """
    search_dir = np.zeros((rank, 1))
    projGradStep = (model_row - gradModel.transpose()) * (
        model_row - (gradModel.transpose() > 0).astype(float)
    )
    wk = np.linalg.norm(model_row - projGradStep)

    # Determine active and free variables
    num_free = 0
    free_indices_tmp = np.zeros((rank,)).astype(int)
    for r in range(rank):
        if (model_row[r] <= np.minimum(epsActSet, wk)) and (gradModel[r] > 0):
            # Variable is not free (belongs to set A or G)
            if model_row[r] != 0:
                # Variable moves according to the gradient (set G).
                search_dir[r] = -gradModel[r]

        else:
            # Variable is free (set F).
            num_free += 1
            free_indices_tmp[num_free - 1] = r

    free_indices = free_indices_tmp[0:num_free]

    # Compute the Hessian for free variables.
    Hessian_free = getHessian(ups_row, Pi, free_indices)
    grad_free = -gradModel[free_indices]

    # Compute the damped Newton search direction over free variables
    # TODO verify this is appropriate representation of matlab's method, s.b. because hessian is square, and addition
    # should ensure full rank, try.catch handles singular matrix
    try:
        search_dir[free_indices] = np.linalg.solve(
            Hessian_free + (mu * np.eye(num_free)), grad_free
        )[:, None]
    except np.linalg.LinAlgError:
        warnings.warn("CP_APR: Damped Hessian is nearly singular\n")
        # TODO: note this may be a typo in matlab see line 1107
        search_dir = -gradModel

    # Calculate expected reduction in the quadratic model of the objective.
    # TODO: double check if left or right multiplication has an speed effect, memory layout
    q = (
        search_dir[free_indices]
        .transpose()
        .dot(Hessian_free + (mu * np.eye(num_free)))
        .dot(search_dir[free_indices])
    )
    pred_red = (search_dir[free_indices].transpose().dot(gradModel[free_indices])) + (
        0.5 * q
    )
    if pred_red > 0:
        warnings.warn("CP_APR: Expected decrease in objective is positive\n")
        search_dir = -gradModel

    return search_dir, pred_red


def tt_linesearch_prowsubprob(
    direction,
    grad,
    model_old,
    step_len,
    step_red,
    max_steps,
    suff_decr,
    isSparse,
    data_row,
    Pi,
    phi_row,
    display_warning,
):
    """
    Perform a line search on a row subproblem

    Parameters
    ----------
    direction:  :class:`numpy.ndarray`
        search direction
    grad:  :class:`numpy.ndarray`
        gradient vector a model_old
    model_old:  :class:`numpy.ndarray`
        current variable values
    step_len: float
        initial step length, which is the maximum possible step length
    step_red: float
        step reduction factor (suggest 1/2)
    max_steps: int
        maximum number of steps to try (suggest 10)
    suff_decr: float
        sufficent decrease for convergence (suggest 1.0e-4)
    isSparse: bool
        sparsity flag for computing the objective
    data_row:  :class:`numpy.ndarray`
        row subproblem data, for computing the objective
    Pi:  :class:`numpy.ndarray`
        Pi matrix, for computing the objective
    phi_row:  :class:`numpy.ndarray`
        1-grad, more accurate if failing over to multiplicative update
    display_warning: bool
        Flag to display warning messages or not

    Returns
    -------
    m_new:  :class:`numpy.ndarray`
        new (improved) model values
    num_evals: int
        number of times objective was evaluated
    f_old: float
        objective value at model_old
    f_1: float
        objective value at model_old + step_len*direction
    f_new: float
        objective value at model_new
    """
    minDescentTol = 1.0e-7
    smallStepTol = 1.0e-7

    stepSize = step_len

    # Evalute the current objective value
    f_old = -tt_loglikelihood_row(isSparse, data_row, model_old, Pi)
    num_evals = 1
    count = 1

    while count <= max_steps:
        # Compute a new step and project it onto the positive orthant.
        model_new = model_old + stepSize * direction
        model_new *= model_new > 0

        # Check that it is a descent direction.
        gDotd = np.sum(grad * (model_new - model_old))
        if (gDotd > 0) or (np.sum(model_new) < minDescentTol):
            # Don't evaluate the objective if not a descent direction
            # or if all the elements of model_new are close to zero
            f_new = np.inf
            if count == 1:
                f_1 = f_new

            stepSize *= step_red
            count += 1
        else:
            # Evaluate objective function at new iterate
            f_new = -ttb.tt_loglikelihood_row(isSparse, data_row, model_new, Pi)
            num_evals += 1
            if count == 1:
                f_1 = f_new

            # Check for sufficient decrease.
            if f_new <= (f_old + suff_decr * gDotd):
                break
            else:
                stepSize *= step_red
                count += 1

    if np.isinf(f_1):
        # Unit step failed; return a value that yields ared =0
        f_1 = f_old

    if ((count >= max_steps) and (f_new > f_old)) or (np.sum(model_new) < smallStepTol):
        # Fall back on a multiplicative update step (scaled steepest descent).
        # Experiments indicate it works better than a unit step in the direction
        # of steepest descent, which would be the following:
        # m_new = m_old - (step_len * grad);     # steepest descent
        # A simple update formula follows, but suffers from round-off error
        # when phi_row is tiny:
        # m_new = m_old - (m_old .* grad);
        # Use this for best accuracy:
        model_new = model_old * phi_row  # multiplicative update

        # Project to the constraints and reevaluate the subproblem objective
        model_new *= model_new > 0
        f_new = -ttb.tt_loglikelihood_row(isSparse, data_row, model_new, Pi)
        num_evals += 1

        # Let the caller know the search direction made no progress.
        f_1 = f_old

        if display_warning:
            warnings.warn(
                "CP_APR: Line search failed, using multiplicative update step"
            )

    return model_new, f_old, f_1, f_new, num_evals


def getHessian(upsilon, Pi, free_indices):
    """
    Return the Hessian for one PDNR row subproblem of Model[n], for just the rows and columns corresponding to the free variables

    Parameters
    ----------
    upsilon: :class:`numpy.ndarray`
        intermediate quantity (upsilon) used for second derivatives
    Pi: :class:`numpy.ndarray`
    free_indices: list

    Returns
    -------
    Hessian: :class:`numpy.ndarray`
        Sub-block of full Hessian identified by free-indices

    """

    num_free = len(free_indices)
    H = np.zeros((num_free, num_free))
    for i in range(num_free):
        for j in range(num_free):
            c = free_indices[i]
            d = free_indices[j]
            val = np.sum(upsilon.transpose() * Pi[:, c] * Pi[:, d])
            H[(i, j), (j, i)] = val
    return H


def tt_loglikelihood_row(isSparse, data_row, model_row, Pi):
    """
    Compute log-likelihood of one row subproblem

    Parameters
    ----------
    isSparse: bool
        Sparsity flag
    data_row: :class:`numpy.ndarray`
        vector of data values
    model_row: :class:`numpy.ndarray`
        vector of model values
    Pi: :class:`numpy.ndarray`

    Notes
    -----
    The row subproblem for a given mode includes one row of matricized tensor
    data (x) and one row of the model (m) in the same matricized mode.
    Then
    (dense case)
    m:  R-length vector
    x:  J-length vector
    Pi: R x J matrix
    (sparse case)
    m:  R-length vector
    x:  p-length vector, where p = nnz in row of matricized data tensor
    Pi: R x p matrix
    F = - (sum_r m_r - sum_j x_j * log (m * Pi_j)
    where Pi_j denotes the j^th column of Pi
    NOTE: Rows of Pi' must sum to one

    Returns
    -------
    loglikelihood: float
        See notes for description

    """
    term1 = -np.sum(model_row)
    if isSparse:
        term2 = np.sum(data_row.transpose() * np.log(model_row.dot(Pi.transpose())))
    else:
        b_pi = model_row.dot(Pi.transpose())
        term2 = 0
        for i in range(len(data_row)):
            if data_row[i] != 0:
                # Define zero times log(anything) to be zero
                term2 += data_row[i] * np.log(b_pi[i])
    loglikelihood = term1 + term2
    return loglikelihood


# PQNR helper functions
def getSearchDirPqnr(
    model_row,
    gradModel,
    epsActSet,
    delta_model,
    delta_grad,
    rho,
    lbfgs_pos,
    iters,
    disp_warn,
):
    """
    Compute the search direction by projecting with L-BFGS.

    Parameters
    ----------
    model_row: :class:`numpy.ndarray`
        current variable values
    gradModel: :class:`numpy.ndarray`
        gradient at model_row
    epsActSet: float
        Bertsekas tolerance for active set determination
    delta_model: :class:`numpy.ndarray`
        L-BFGS array of variable deltas
    delta_grad: :class:`numpy.ndarray`
        L-BFGS array of gradient deltas
    rho:
    lbfgs_pos: int
        pointer into L-BFGS arrays
    iters:
    disp_warn: bool

    Returns
    -------
    direction: :class:`numpy.ndarray`
        Search direction based on current L-BFGS and grad

    Notes
    -----
    Adapted from MATLAB code of Dongmin Kim and Suvrit Sra written in 2008.
    Modified extensively to solve row subproblems and use a better linesearch;  for details see
    REFERENCE: Samantha Hansen, Todd Plantenga, Tamara G. Kolda.
    Newton-Based Optimization for Nonnegative Tensor Factorizations,
    arXiv:1304.4964 [math.NA], April 2013,
    URL: http://arxiv.org/abs/1304.4964. Submitted for publication.

    """

    lbfgsSize = delta_model.shape[1]

    # Determine active and free variables.
    # TODO: is the bellow relevant?
    # If epsActSet is zero, then the following works:
    # fixedVars = find((m_row == 0) & (grad' > 0));
    # For the general case this works but is less clear and assumes m_row > 0:
    # fixedVars = find((grad' > 0) & (m_row <= min(epsActSet,grad')));
    projGradStep = (model_row - gradModel.transpose()) * (
        model_row - (gradModel.transpose() > 0).astype(float)
    )
    wk = np.linalg.norm(model_row - projGradStep)
    fixedVars = np.logical_and(gradModel > 0, (model_row <= np.minimum(epsActSet, wk)))

    direction = -gradModel
    direction[fixedVars] = 0

    if delta_model[:, lbfgs_pos].transpose().dot(delta_grad[:, lbfgs_pos]) == 0.0:
        # Cannot proceed with this L-BFGS data; most likely the iteration has converged, so this is rarely seen.
        if disp_warn:
            warnings.warn("WARNING: L-BFGS update is orthogonal, using gradient")
        return direction

    alpha = np.ones((lbfgsSize,))
    k = lbfgs_pos

    # Perform an L-BFGS two-loop recursion to compute the search direction.
    for i in range(np.minimum(iters, lbfgsSize)):
        alpha[k] = rho[k] * (delta_model[:, k].transpose().dot(direction))
        direction -= alpha[k] * (delta_grad[:, k])
        # TODO check mod
        k = (
            lbfgsSize - np.mod(1 - k, lbfgsSize) - 1
        )  # -1 accounts for numpy indexing starting at 0 not 1

    coef = (
        1
        / rho[lbfgs_pos]
        / delta_grad[:, lbfgs_pos].transpose().dot(delta_grad[:, lbfgs_pos])
    )
    direction *= coef

    for i in range(np.minimum(iters, lbfgsSize)):
        k = np.mod(k, lbfgsSize)  # + 1
        b = rho[k] * (delta_grad[:, k].transpose().dot(direction))
        direction += (alpha[k] - b) * (delta_model[:, k])

    direction[fixedVars] = 0

    return direction


def calc_grad(isSparse, Pi, eps_div_zero, data_row, model_row):
    """
    Compute the gradient for a PQNR row subproblem

    Parameters
    ----------
    isSparse: bool
    Pi: :class:`numpy.ndarray`
    eps_div_zero: float
    data_row: :class:`numpy.ndarray`
    model_row: :class:`numpy.ndarray`

    Returns
    -------
    phi_row: :class:`numpy.ndarray`
    grad_row: :class:`numpy.ndarray`

    """
    # TODO: note this is duplicated exactly from calc_partials, should combine to one function
    if isSparse:
        data_row = data_row.transpose()[0]
    v = model_row.dot(Pi.transpose())
    w = data_row / np.maximum(v, eps_div_zero)
    phi_row = w.dot(Pi)
    # print("V: {}, W :{}".format(v,w))
    # u = v**2
    # ups_row = data_row.transpose() / np.maximum(u, epsilon)
    grad_row = (np.ones(phi_row.shape) - phi_row).transpose()
    return grad_row, phi_row


# Mu helper functions
def calculatePi(Data, Model, rank, factorIndex, ndims):
    """
    Helper function to calculate Pi matrix
    # TODO verify what pi is

    Parameters
    ----------
    Data: :class:`pyttb.sptensor` or :class:`pyttb.tensor`
    Model: :class:`pyttb.ktensor`
    rank: int
    factorIndex: int
    ndims: int

    Returns
    -------
    Pi: :class:`numpy.ndarray`
    """
    if isinstance(Data, ttb.sptensor):
        Pi = np.ones((Data.nnz, rank))
        for i in np.setdiff1d(np.arange(ndims), factorIndex).astype(int):
            Pi *= Model[i][Data.subs[:, i], :]
    else:
        Pi = ttb.khatrirao(
            Model.factor_matrices[:factorIndex]
            + Model.factor_matrices[factorIndex + 1 :],
            reverse=True,
        )

    return Pi


def calculatePhi(Data, Model, rank, factorIndex, Pi, epsilon):
    """

    Parameters
    ----------
    Data: :class:`pyttb.sptensor` or :class:`pyttb.tensor`
    Model: :class:`pyttb.ktensor`
    rank: int
    factorIndex: int
    Pi: :class:`numpy.ndarray`
    epsilon: float

    Returns
    -------

    """
    if isinstance(Data, ttb.sptensor):
        Phi = -np.ones((Data.shape[factorIndex], rank))
        xsubs = Data.subs[:, factorIndex]
        v = np.sum(Model.factor_matrices[factorIndex][xsubs, :] * Pi, axis=1)
        wvals = Data.vals / np.maximum(v, epsilon)[:, None]
        for r in range(rank):
            Yr = accumarray(
                xsubs,
                np.squeeze(wvals * Pi[:, r][:, None]),
                size=Data.shape[factorIndex],
            )
            Phi[:, r] = Yr
    else:
        Xn = ttb.tt_to_dense_matrix(Data, factorIndex)
        V = Model.factor_matrices[factorIndex].dot(Pi.transpose())
        W = Xn / np.maximum(V, epsilon)
        Y = W.dot(Pi)
        Phi = Y

    return Phi


def tt_loglikelihood(Data, Model):
    """
    Compute log-likelihood of data with model.

    Parameters
    ----------
    Data: :class:`pyttb.sptensor` or :class:`pyttb.tensor`
    Model: :class:`pyttb.ktensor`

    Returns
    -------
    loglikelihood: float
        - (sum_i m_i - x_i * log_i) where i is a multiindex across all tensor dimensions.

    Notes
    -----
    We define for any x 0*log(x)=0, such that if our true data is 0 the loglikelihood is the value of the model.

    """
    N = Data.ndims

    assert isinstance(Model, ttb.ktensor), "Model must be a ktensor"

    Model.normalize(weight_factor=0, normtype=1)
    if isinstance(Data, ttb.sptensor):
        xsubs = Data.subs
        A = Model.factor_matrices[0][xsubs[:, 0], :]
        for n in range(1, N):
            A *= Model.factor_matrices[n][xsubs[:, n], :]
        return np.sum(Data.vals * np.log(np.sum(A, axis=1))[:, None]) - np.sum(
            Model.factor_matrices[0]
        )
    else:
        dX = ttb.tt_to_dense_matrix(Data, 1)
        dM = ttb.tt_to_dense_matrix(Model, 1)
        f = 0
        for i in range(dX.shape[0]):
            for j in range(dX.shape[1]):
                if dX[i, j] == 0:
                    pass
                else:
                    f += dX[i, j] * np.log(dM[i, j])

        f -= np.sum(Model.factor_matrices[0])
        return f


def vectorizeForMu(matrix):
    """
    Helper Function to unravel matrix into vector

    Parameters
    ----------
    matrix: :class:`numpy.ndarray`

    Returns
    -------
    matrix: :class:`numpy.ndarray`
        len(matrix.shape)==1
    """
    return matrix.ravel()



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.

import numpy as np

import pyttb as ttb

from .pyttb_utils import *


class tenmat(object):
    """
    TENMAT Store tensor as a matrix.

    """

    def __init__(self):
        """
        TENSOR Create empty tensor.
        """

        # Case 0a: Empty Contructor
        self.tshape = ()
        self.rindices = np.array([])
        self.cindices = np.array([])
        self.data = np.array([])

    @classmethod
    def from_data(cls, data, rdims, cdims=None, tshape=None):
        # CONVERT A MULTIDIMENSIONAL ARRAY

        # Verify that data is a numeric numpy.ndarray
        if not isinstance(data, np.ndarray) or not issubclass(
            data.dtype.type, np.number
        ):
            assert False, "First argument must be a numeric numpy.ndarray."

        # data is empty, return empty tenmat unless rdims, cdims, or tshape are not empty
        if data.size == 0:
            if not rdims.size == 0 or not cdims.size == 0 or not tshape == ():
                assert (
                    False
                ), "When data is empty, rdims, cdims, and tshape must also be empty."
            else:
                return cls()

        # data is 1d array, must convert to 2d array for tenmat
        if len(data.shape) == 1:
            if tshape is None:
                assert False, "tshape must be specified when data is 1d array."
            else:
                # make data a 2d array with shape (1, data.shape[0]), i.e., a row vector
                data = np.reshape(data.copy(), (1, data.shape[0]), order="F")

        # data is ndarray and only rdims is specified
        if cdims is None:
            return ttb.tenmat.from_tensor_type(ttb.tensor.from_data(data), rdims)

        # use data.shape for tshape if not provided
        if tshape is None:
            tshape = data.shape
        elif not isinstance(tshape, tuple):
            assert False, "tshape must be a tuple."

        # check that data.shape and tshape agree
        if not np.prod(data.shape) == np.prod(tshape):
            assert (
                False
            ), "Incorrect dimensions specified: products of data.shape and tuple do not match"

        # check that data.shape and product of dimensions agree
        if not np.prod(np.array(tshape)[rdims]) * np.prod(
            np.array(tshape)[cdims]
        ) == np.prod(data.shape):
            assert (
                False
            ), "data.shape does not match shape specified by rdims, cdims, and tshape."

        return ttb.tenmat.from_tensor_type(
            ttb.tensor.from_data(data, tshape), rdims, cdims
        )

    @classmethod
    def from_tensor_type(cls, source, rdims=None, cdims=None, cdims_cyclic=None):
        # Case 0b: Copy Contructor
        if isinstance(source, tenmat):
            # Create tenmat
            tenmatInstance = cls()
            tenmatInstance.tshape = source.tshape
            tenmatInstance.rindices = source.rindices.copy()
            tenmatInstance.cindices = source.cindices.copy()
            tenmatInstance.data = source.data.copy()
            return tenmatInstance

        # Case III: Convert a tensor to a tenmat
        if isinstance(source, ttb.tensor):
            n = source.ndims
            alldims = np.array([range(n)])
            tshape = source.shape

            # Verify inputs
            if rdims is None and cdims is None:
                assert False, "Either rdims or cdims or both must be specified."
            if rdims is not None and not sum(np.in1d(rdims, alldims)) == len(rdims):
                assert False, "Values in rdims must be in [0, source.ndims]."
            if cdims is not None and not sum(np.in1d(cdims, alldims)) == len(cdims):
                assert False, "Values in cdims must be in [0, source.ndims]."

            if rdims is not None and cdims is None:
                # Single row mapping
                if len(rdims) == 1 and cdims_cyclic is not None:
                    if cdims_cyclic == "fc":
                        # cdims = [rdims+1:n, 1:rdims-1];
                        cdims = np.array(
                            [i for i in range(rdims[0] + 1, n)]
                            + [i for i in range(rdims[0])]
                        )
                    elif cdims_cyclic == "bc":
                        # cdims = [rdims-1:-1:1, n:-1:rdims+1];
                        cdims = np.array(
                            [i for i in range(rdims[0] - 1, -1, -1)]
                            + [i for i in range(n - 1, rdims[0], -1)]
                        )
                    else:
                        assert (
                            False
                        ), 'Unrecognized value for cdims_cyclic pattern, must be "fc" or "bc".'

                else:
                    # Multiple row mapping
                    cdims = np.setdiff1d(alldims, rdims)

            elif rdims is None and cdims is not None:
                rdims = np.setdiff1d(alldims, cdims)

            # if rdims or cdims is empty, hstack will output an array of float not int
            if rdims.size == 0:
                dims = cdims.copy()
            elif cdims.size == 0:
                dims = rdims.copy()
            else:
                dims = np.hstack([rdims, cdims])
            if not len(dims) == n or not (alldims == np.sort(dims)).all():
                assert (
                    False
                ), "Incorrect specification of dimensions, the sorted concatenation of rdims and cdims must be range(source.ndims)."

            rprod = 1 if rdims.size == 0 else np.prod(np.array(tshape)[rdims])
            cprod = 1 if cdims.size == 0 else np.prod(np.array(tshape)[cdims])
            data = np.reshape(source.permute(dims).data, (rprod, cprod), order="F")

            # Create tenmat
            tenmatInstance = cls()
            tenmatInstance.tshape = tshape
            tenmatInstance.rindices = rdims.copy()
            tenmatInstance.cindices = cdims.copy()
            tenmatInstance.data = data.copy()
            return tenmatInstance

    def ctranspose(self):
        """
        Complex conjugate transpose for tenmat.

        Parameters
        ----------

        Returns
        -------
        :class:`pyttb.tenmat`
        """

        tenmatInstance = tenmat()
        tenmatInstance.rindices = self.cindices.copy()
        tenmatInstance.cindices = self.rindices.copy()
        tenmatInstance.tshape = self.tshape
        tenmatInstance.data = self.data.conj().T.copy()
        return tenmatInstance

    def double(self):
        """
        Convert tenmat to an array of doubles

        Returns
        -------
        :class:`numpy.ndarray`
            copy of tenmat data
        """
        return self.data.astype(np.float_).copy()

    def end(self, k):
        """
        Last index of indexing expression for tenmat

        Parameters
        ----------
        k: int
            dimension for subscripted indexing

        Returns
        -------
        int: index
        """

        return self.data.shape[k] - 1

    @property
    def ndims(self):
        """
        Return the number of dimensions of a tenmat

        Returns
        -------
        int
        """
        return len(self.shape)

    def norm(self):
        """
        Frobenius norm of a tenmat.

        Parameters
        ----------

        Returns
        -------
        Returns
        -------
        float
        """
        # default of np.linalg.norm is to vectorize the data and compute the vector norm, which is equivalent to
        # the Frobenius norm for multidimensional arrays. However, the argument 'fro' only workks for 1-D and 2-D
        # arrays currently.
        return np.linalg.norm(self.data)

    @property
    def shape(self):
        """
        Return the shape of a tenmat

        Returns
        -------
        tuple
        """
        if self.data.shape == (0,):
            return ()
        else:
            return self.data.shape

    def __setitem__(self, key, value):
        """
        SUBSASGN Subscripted assignment for a tensor.
        """
        self.data[key] = value
        return

    def __getitem__(self, item):
        """
        SUBSREF Subscripted reference for tenmat.

        Parameters
        ----------
        item:

        Returns
        -------
        :class:`numpy.ndarray`, float, int
        """
        return self.data[item]

    def __mul__(self, other):
        """
        Multiplies two tenmat objects.

        Parameters
        ----------
        other: :class:`pyttb.tenmat`

        Returns
        -------
        :class:`pyttb.tenmat`
        """
        # One argument is a scalar
        if np.isscalar(other):
            Z = ttb.tenmat.from_tensor_type(self)
            Z.data = Z.data * other
            return Z
        elif isinstance(other, tenmat):
            # Check that data shapes are compatible
            if not self.shape[1] == other.shape[0]:
                assert (
                    False
                ), "tenmat shape mismatch: number or columns of left operand must match number of rows of right operand."

            tshape = tuple(
                np.hstack(
                    (
                        np.array(self.tshape)[self.rindices],
                        np.array(other.tshape)[other.cindices],
                    )
                )
            )

            if tshape == ():
                return (self.data @ other.data)[0, 0]
            else:
                tenmatInstance = tenmat()
                tenmatInstance.tshape = tshape
                tenmatInstance.rindices = np.arange(len(self.rindices))
                tenmatInstance.cindices = np.arange(len(other.cindices)) + len(
                    self.rindices
                )
                tenmatInstance.data = self.data @ other.data
                return tenmatInstance
        else:
            assert (
                False
            ), "tenmat multiplication only valid with scalar or tenmat objects."

    def __rmul__(self, other):
        """
        Multiplies two tenmat objects.

        Parameters
        ----------
        other: :class:`pyttb.tenmat`

        Returns
        -------
        :class:`pyttb.tenmat`
        """
        return self.__mul__(other)

    def __add__(self, other):
        pass
        """
        Binary addition (+) for tenmats

        Parameters
        ----------
        other: :class:`pyttb.tenmat`, float, int

        Returns
        -------
        :class:`pyttb.tenmat`
        """

        # One argument is a scalar
        if np.isscalar(other):
            Z = ttb.tenmat.from_tensor_type(self)
            Z.data = Z.data + other
            return Z
        elif isinstance(other, tenmat):
            # Check that data shapes agree
            if not self.shape == other.shape:
                assert False, "tenmat shape mismatch."

            Z = ttb.tenmat.from_tensor_type(self)
            Z.data = Z.data + other.data
            return Z
        else:
            assert False, "tenmat addition only valid with scalar or tenmat objects."

    def __radd__(self, other):
        """
        Reverse binary addition (+) for tenmats

        Parameters
        ----------
        other: :class:`pyttb.tenmat`, float, int

        Returns
        -------
        :class:`pyttb.tenmat`
        """
        return self.__add__(other)

    def __sub__(self, other):
        pass
        """
        Binary subtraction (-) for tenmats

        Parameters
        ----------
        other: :class:`pyttb.tenmat`, float, int

        Returns
        -------
        :class:`pyttb.tenmat`
        """

        # One argument is a scalar
        if np.isscalar(other):
            Z = ttb.tenmat.from_tensor_type(self)
            Z.data = Z.data - other
            return Z
        elif isinstance(other, tenmat):
            # Check that data shapes agree
            if not self.shape == other.shape:
                assert False, "tenmat shape mismatch."

            Z = ttb.tenmat.from_tensor_type(self)
            Z.data = Z.data - other.data
            return Z
        else:
            assert False, "tenmat subtraction only valid with scalar or tenmat objects."

    def __rsub__(self, other):
        """
        Reverse binary subtraction (-) for tenmats

        Parameters
        ----------
        other: :class:`pyttb.tenmat`, float, int

        Returns
        -------
        :class:`pyttb.tenmat`
        """

        # One argument is a scalar
        if np.isscalar(other):
            Z = ttb.tenmat.from_tensor_type(self)
            Z.data = other - Z.data
            return Z
        elif isinstance(other, tenmat):
            # Check that data shapes agree
            if not self.shape == other.shape:
                assert False, "tenmat shape mismatch."

            Z = ttb.tenmat.from_tensor_type(self)
            Z.data = other.data - Z.data
            return Z
        else:
            assert False, "tenmat subtraction only valid with scalar or tenmat objects."

    def __pos__(self):
        """
        Unary plus (+) for tenmats

        Returns
        -------
        :class:`pyttb.tenmat`
            copy of tenmat
        """

        T = ttb.tenmat.from_tensor_type(self)

        return T

    def __neg__(self):
        """
        Unary minus (-) for tenmats

        Returns
        -------
        :class:`pyttb.tenmat`
            copy of tenmat
        """

        T = ttb.tenmat.from_tensor_type(self)
        T.data = -1 * T.data

        return T

    def __repr__(self):
        """
        String representation of a tenmat.

        Returns
        -------
        str
            Contains the shape, row indices (rindices), column indices (cindices) and data as strings on different lines.
        """
        s = ""
        s += "matrix corresponding to a tensor of shape "
        if self.data.size == 0:
            s += str(self.shape)
        else:
            s += (" x ").join([str(int(d)) for d in self.tshape])
        s += "\n"

        s += "rindices = "
        s += "[ " + (", ").join([str(int(d)) for d in self.rindices]) + " ] "
        s += "(modes of tensor corresponding to rows)\n"

        s += "cindices = "
        s += "[ " + (", ").join([str(int(d)) for d in self.cindices]) + " ] "
        s += "(modes of tensor corresponding to columns)\n"

        if self.data.size == 0:
            s += "data = []\n"
        else:
            s += "data[:, :] = \n"
            s += str(self.data)
            s += "\n"

        return s

    __str__ = __repr__



# Copyright 2022 National Technology & Engineering Solutions of Sandia,
# LLC (NTESS). Under the terms of Contract DE-NA0003525 with NTESS, the
# U.S. Government retains certain rights in this software.
"""Dense Tensor Implementation"""
from __future__ import annotations

import warnings
from itertools import permutations
from math import factorial
from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
import scipy.sparse.linalg
from numpy_groupies import aggregate as accumarray

import pyttb as ttb
from pyttb.pyttb_utils import tt_dimscheck, tt_ind2sub


class tensor:
    """
    TENSOR Class for dense tensors.
    """

    data: np.ndarray
    shape: Tuple

    def __init__(self):
        """
        TENSOR Create empty tensor.
        """

        # EMPTY / DEFAULT CONSTRUCTOR
        self.data = np.array([])
        self.shape = ()

    @classmethod
    def from_data(
        cls, data: np.ndarray, shape: Optional[Tuple[int, ...]] = None
    ) -> tensor:
        """
        Creates a tensor from explicit description. Note that 1D tensors (i.e.,
        when len(shape)==1) contains a data array that follow the Numpy convention
        of being a row vector, which is different than in the Matlab Tensor Toolbox.

        Parameters
        ----------
        data: Tensor source data
        shape: Shape of resulting tensor if not the same as data shape

        Returns
        -------
        Constructed tensor

        Example
        -------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> Y = ttb.tensor.from_data(np.ones((2,2)), shape=(4,1))
        """
        # CONVERT A MULTIDIMENSIONAL ARRAY
        if not issubclass(data.dtype.type, np.number) and not issubclass(
            data.dtype.type, np.bool_
        ):
            assert False, "First argument must be a multidimensional array."

        # Create or check second argument
        if shape is None:
            shape = data.shape
        else:
            if not isinstance(shape, tuple):
                assert False, "Second argument must be a tuple."

        # Make sure the number of elements matches what's been specified
        if len(shape) == 0:
            if data.size > 0:
                assert False, "Empty tensor cannot contain any elements"

        elif np.prod(shape) != data.size:
            assert (
                False
            ), "TTB:WrongSize, Size of data does not match specified size of tensor"

        # Make sure the data is indeed the right shape
        if data.size > 0 and len(shape) > 0:
            # reshaping using Fortran ordering to match Matlab conventions
            data = np.reshape(data, np.array(shape), order="F")

        # Create the tensor
        tensorInstance = cls()
        tensorInstance.data = data.copy()
        tensorInstance.shape = shape
        return tensorInstance

    @classmethod
    def from_tensor_type(
        cls, source: Union[ttb.sptensor, tensor, ttb.ktensor, ttb.tenmat]
    ) -> tensor:
        """
        Converts other tensor types into a dense tensor

        Parameters
        ----------
        source: Tensor type to create dense tensor from

        Returns
        -------
        Constructed tensor

        Example
        -------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> Y = ttb.tensor.from_tensor_type(X)
        """
        # CONVERSION/COPY CONSTRUCTORS
        if isinstance(source, tensor):
            # COPY CONSTRUCTOR
            return cls.from_data(source.data.copy(), source.shape)
        if isinstance(
            source,
            (
                ttb.ktensor,
                ttb.ttensor,
                ttb.sptensor,
                ttb.sumtensor,
                ttb.symtensor,
                ttb.symktensor,
            ),
        ):
            # CONVERSION
            t = source.full()
            return cls.from_data(t.data.copy(), t.shape)
        if isinstance(source, ttb.tenmat):
            # RESHAPE TENSOR-AS-MATRIX
            # Here we just reverse what was done in the tenmat constructor.
            # First we reshape the data to be an MDA, then we un-permute
            # it using ipermute.
            shape = source.tshape
            order = np.hstack([source.rindices, source.cindices])
            data = np.reshape(source.data.copy(), np.array(shape)[order], order="F")
            if order.size > 1:
                data = np.transpose(data, np.argsort(order))
            return cls.from_data(data, shape)
        raise ValueError(f"Unsupported type for tensor source, received {type(source)}")

    @classmethod
    def from_function(
        cls,
        function_handle: Callable[[Tuple[int, ...]], np.ndarray],
        shape: Tuple[int, ...],
    ) -> tensor:
        """
        Creates a tensor from a function handle and size

        Parameters
        ----------
        function_handle: Function to generate data to construct tensor
        shape: Shape of resulting tensor

        Returns
        -------
        Constructed tensor

        Example
        -------
        >>> X = ttb.tensor.from_function(lambda a_shape: np.ones(a_shape), (2,2))
        """
        # FUNCTION HANDLE AND SIZE

        # Check size
        if not isinstance(shape, tuple):
            assert False, "TTB:BadInput, Shape must be a tuple"

        # Generate data
        data = function_handle(shape)

        # Create the tensor
        return cls.from_data(data, shape)

    def collapse(
        self,
        dims: Optional[np.ndarray] = None,
        fun: Callable[[np.ndarray], Union[float, np.ndarray]] = np.sum,
    ) -> Union[float, np.ndarray, tensor]:
        """
        Collapse tensor along specified dimensions.

        Parameters
        ----------
        dims: Dimensions to collapse
        fun: Method used to collapse dimensions

        Returns
        -------
        Collapsed value

        Example
        -------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> X.collapse()
        4.0
        >>> X.collapse(np.arange(X.ndims), sum)
        4.0
        """
        if self.data.size == 0:
            return np.array([])

        if dims is None:
            dims = np.arange(0, self.ndims)

        if dims.size == 0:
            return ttb.tensor.from_tensor_type(self)

        dims, _ = tt_dimscheck(self.ndims, dims=dims)
        remdims = np.setdiff1d(np.arange(0, self.ndims), dims)

        # Check for the case where we accumulate over *all* dimensions
        if remdims.size == 0:
            return fun(self.data.flatten("F"))

        ## Calculate the shape of the result
        newshape = tuple(np.array(self.shape)[remdims])

        ## Convert to a matrix where each row is going to be collapsed
        A = ttb.tenmat.from_data(self.data, remdims, dims).double()

        ## Apply the collapse function
        B = np.zeros((A.shape[0], 1))
        for i in range(0, A.shape[0]):
            B[i] = fun(A[i, :])

        ## Form and return the final result
        return ttb.tensor.from_data(B, newshape)

    def contract(self, i: int, j: int) -> Union[np.ndarray, tensor]:
        """
        Contract tensor along two dimensions (array trace).

        Parameters
        ----------
        i: First dimension
        j: Second dimension

        Returns
        -------
        Contracted tensor

        Example
        -------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> X.contract(0, 1)
        2.0
        """
        if self.shape[i] != self.shape[j]:
            assert False, "Must contract along equally sized dimensions"

        if i == j:
            assert False, "Must contract along two different dimensions"

        # Easy case - returns a scalar
        if self.ndims == 2:
            return np.trace(self.data)

        # Remaining dimensions after trace
        remdims = np.setdiff1d(np.arange(0, self.ndims), np.array([i, j])).astype(int)

        # Size for return
        newsize = tuple(np.array(self.shape)[remdims])

        # Total size of remainder
        m = np.prod(newsize)

        # Number of items to add for trace
        n = self.shape[i]

        # Permute trace dimensions to the end
        x = self.permute(np.concatenate((remdims, np.array([i, j]))))

        # Reshape data to be 3D
        data = np.reshape(x.data, (m, n, n), order="F")

        # Add diagonal entries for each slice
        newdata = np.zeros((m, 1))
        for idx in range(0, n):
            newdata += data[:, idx, idx][:, None]

        # Reshape result
        if np.prod(newsize) > 1:
            newdata = np.reshape(newdata, newsize, order="F")

        return ttb.tensor.from_data(newdata, newsize)

    def double(self) -> np.ndarray:
        """
        Convert tensor to an array of doubles

        Returns
        -------
        Copy of tensor data

        Example
        -------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> X.double()
        array([[1., 1.],
               [1., 1.]])
        """
        return self.data.astype(np.float_).copy()

    def exp(self) -> tensor:
        """
        Exponential of the elements of tensor

        Returns
        -------
        Copy of tensor data element-wise raised to exponential

        Examples
        --------
        >>> tensor1 = ttb.tensor.from_data(np.array([[1, 2], [3, 4]]))
        >>> tensor1.exp().data  # doctest: +ELLIPSIS
        array([[ 2.7182...,  7.3890... ],
               [20.0855..., 54.5981...]])
        """
        return ttb.tensor.from_data(np.exp(self.data))

    def end(self, k: Optional[int] = None) -> int:
        """
        Last index of indexing expression for tensor

        Parameters
        ----------
        k: dimension for subscripted indexing

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> X.end()  # linear indexing
        3
        >>> X.end(0)
        1
        """

        if k is not None:  # Subscripted indexing
            return self.shape[k] - 1
        # For linear indexing
        return np.prod(self.shape) - 1

    def find(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        FIND Find subscripts of nonzero elements in a tensor.

        S, V = FIND(X) returns the subscripts of the nonzero values in X and a column
        vector of the values.

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.zeros((3,4,2)))
        >>> larger_entries = X > 0.5
        >>> subs, vals = larger_entries.find()

        See Also
        --------
        TENSOR/SUBSREF, TENSOR/SUBSASGN

        Returns
        -------
        Subscripts and values for non-zero entries
        """
        idx = np.nonzero(np.ravel(self.data, order="F"))[0]
        subs = ttb.tt_ind2sub(self.shape, idx)
        vals = self.data[tuple(subs.T)][:, None]
        return subs, vals

    def full(self) -> tensor:
        """
        Convert dense tensor to dense tensor.

        Returns
        -------
        Deep copy
        """
        return ttb.tensor.from_data(self.data)

    def innerprod(self, other: Union[tensor, ttb.sptensor, ttb.ktensor]) -> float:
        """
        Efficient inner product with a tensor

        Parameters
        ----------
        other: Tensor type to take an innerproduct with

        Examples
        --------
        >>> tensor1 = ttb.tensor.from_data(np.array([[1, 2], [3, 4]]))
        >>> tensor1.innerprod(tensor1)
        30
        """
        if isinstance(other, ttb.tensor):
            if self.shape != other.shape:
                assert False, "Inner product must be between tensors of the same size"
            x = np.reshape(self.data, (self.data.size,), order="F")
            y = np.reshape(other.data, (other.data.size,), order="F")
            return x.dot(y)
        if isinstance(other, (ttb.ktensor, ttb.sptensor, ttb.ttensor)):
            # Reverse arguments and call specializer code
            return other.innerprod(self)
        assert False, "Inner product between tensor and that class is not supported"

    def isequal(self, other: Union[tensor, ttb.sptensor]) -> bool:
        """
        Exact equality for tensors

        Parameters
        ----------
        other: Tensor to compare against

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> Y = ttb.tensor.from_data(np.zeros((2,2)))
        >>> X.isequal(Y)
        False
        """
        if isinstance(other, ttb.tensor):
            return bool(np.all(self.data == other.data))
        if isinstance(other, ttb.sptensor):
            return bool(np.all(self.data == other.full().data))
        return False

    # TODO: We should probably always return details and let caller drop them
    # pylint: disable=too-many-branches, too-many-locals
    def issymmetric(
        self,
        grps: Optional[np.ndarray] = None,
        version: Optional[Any] = None,
        return_details: bool = False,
    ) -> Union[bool, Tuple[bool, np.ndarray, np.ndarray]]:
        """
        Determine if a dense tensor is symmetric in specified modes.

        Parameters
        ----------
        grps: Modes to check for symmetry
        version: Flag
            Any non-None value will call the non-default old version
        return_details: Flag to return symmetry details in addition to bool

        Returns
        -------
        If symmetric in modes, optionally all differences and permutations

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> X.issymmetric()
        True
        >>> X.issymmetric(grps=np.arange(X.ndims))
        True
        >>> is_sym, diffs, perms = \
            X.issymmetric(grps=np.arange(X.ndims), version=1, return_details=True)
        >>> print(f"Tensor is symmetric: {is_sym}")
        Tensor is symmetric: True
        >>> print(f"Differences in modes: {diffs}")
        Differences in modes: [[0.]
         [0.]]
        >>> print(f"Permutations: {perms}")
        Permutations: [[0. 1.]
         [1. 0.]]
        """
        n = self.ndims
        sz = np.array(self.shape)

        if grps is None:
            grps = np.arange(0, n)

        if len(grps.shape) == 1:
            grps = np.array([grps])

        # Substantially different routines are called depending on whether the user
        # requests the permutation information. If permutation is required
        # (or requested) the algorithm is much slower
        if version is None:  # pylint:disable=no-else-return
            # Use new algorithm
            for thisgrp in grps:
                # Check tensor dimensions first
                if not np.all(sz[thisgrp[0]] == sz[thisgrp]):
                    return False

                # Construct matrix ind where each row is the multi-index for one
                # element of X
                idx = tt_ind2sub(self.shape, np.arange(0, self.data.size))

                # Find reference index for every element in the tensor - this
                # is to its index in the symmetrized tensor. This puts every
                # element into a 'class' of entries that will be the same under
                # symmetry.
                classidx = idx
                classidx[:, thisgrp] = np.sort(idx[:, thisgrp], axis=1)

                # Compare each element to its class exemplar
                if np.any(self.data.ravel() != self.data[tuple(classidx.transpose())]):
                    return False

            # We survived all the tests!
            return True

        # Use the older algorithm
        else:
            # Check tensor dimensions for compatibility with symmetrization
            for dims in grps:
                for j in dims[1:]:
                    if sz[j] != sz[dims[0]]:
                        return False

            # Check actual symmetry
            cnt = sum(factorial(len(x)) for x in grps)
            all_diffs = np.zeros((cnt, 1))
            all_perms = np.zeros((cnt, n))
            for a_group in grps:
                # Compute the permutations for this group of symmetries
                for p_idx, perm in enumerate(permutations(a_group)):
                    all_perms[p_idx, :] = perm

                    # Do the permutation and record the difference.
                    Y = self.permute(np.array(perm))
                    if np.array_equal(self.data, Y.data):
                        all_diffs[p_idx] = 0
                    else:
                        all_diffs[p_idx] = np.max(
                            np.abs(self.data.ravel() - Y.data.ravel())
                        )

            if return_details is False:
                return bool((all_diffs == 0).all())
            return bool((all_diffs == 0).all()), all_diffs, all_perms

    def logical_and(self, B: Union[float, tensor]) -> tensor:
        """
        Logical and for tensors

        Parameters
        ----------
        B: Value to and against self

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2), dtype=bool))
        >>> X.logical_and(X).collapse()  # All true
        4
        """

        def logical_and(x, y):
            return np.logical_and(x, y)

        return ttb.tt_tenfun(logical_and, self, B)

    def logical_not(self) -> tensor:
        """
        Logical Not For Tensors

        Returns
        -------
        Negated tensor

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2), dtype=bool))
        >>> X.logical_not().collapse()  # All false
        0
        """
        return ttb.tensor.from_data(np.logical_not(self.data))

    def logical_or(self, other: Union[float, tensor]) -> tensor:
        """
        Logical or for tensors

        Parameters
        ----------
        other: Value to perform or against

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2), dtype=bool))
        >>> X.logical_or(X.logical_not()).collapse()  # All true
        4
        """

        def tensor_or(x, y):
            return np.logical_or(x, y)

        return ttb.tt_tenfun(tensor_or, self, other)

    def logical_xor(self, other: Union[float, tensor]) -> tensor:
        """
        Logical xor for tensors

        Parameters
        ----------
        other: Value to perform xor against

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2), dtype=bool))
        >>> X.logical_xor(X.logical_not()).collapse()  # All true
        4
        """

        def tensor_xor(x, y):
            return np.logical_xor(x, y)

        return ttb.tt_tenfun(tensor_xor, self, other)

    def mask(self, W: tensor) -> np.ndarray:
        """
        Extract non-zero values at locations specified by mask tensor

        Parameters
        ----------
        W: Mask tensor

        Returns
        -------
        Extracted values

        Examples
        --------
        >>> W = ttb.tensor.from_data(np.ones((2,2)))
        >>> tensor1 = ttb.tensor.from_data(np.array([[1, 2], [3, 4]]))
        >>> tensor1.mask(W)
        array([1, 3, 2, 4])
        """
        # Error checking
        if np.any(np.array(W.shape) > np.array(self.shape)):
            assert False, "Mask cannot be bigger than the data tensor"

        # Extract locations of nonzeros in W
        wsubs, _ = W.find()

        # Extract those non-zero values
        return self.data[tuple(wsubs.transpose())]

    # pylint: disable=too-many-branches
    def mttkrp(self, U: Union[ttb.ktensor, List[np.ndarray]], n: int) -> np.ndarray:
        """
        Matricized tensor times Khatri-Rao product

        Parameters
        ----------
        U: Matrices to create the Khatri-Rao product
        n: Mode to matricize tensor in

        Returns
        -------
        Matrix product

        Example
        -------
        >>> tensor1 = ttb.tensor.from_data(np.ones((2,2,2)))
        >>> matrices = [np.ones((2,2))] * 3
        >>> tensor1.mttkrp(matrices, 2)
        array([[4., 4.],
               [4., 4.]])
        """

        # check that we have a tensor that can perform mttkrp
        if self.ndims < 2:
            assert False, "MTTKRP is invalid for tensors with fewer than 2 dimensions"

        # extract the list of factor matrices if given a ktensor
        if isinstance(U, ttb.ktensor):
            U = ttb.ktensor.from_tensor_type(U)
            if n == 0:
                U.redistribute(1)
            else:
                U.redistribute(0)
            # Extract the factor matrices
            U = U.factor_matrices

        # check that we have a list (or list extracted from a ktensor)
        if not isinstance(U, list):
            assert False, "Second argument should be a list of arrays or a ktensor"

        # check that list is the correct length
        if len(U) != self.ndims:
            assert False, "Second argument contains the wrong number of arrays"

        if n == 0:
            R = U[1].shape[1]
        else:
            R = U[0].shape[1]

        # check that the dimensions match
        for i in range(self.ndims):
            if i == n:
                continue
            if U[i].shape[0] != self.shape[i]:
                assert False, f"Entry {i} of list of arrays is wrong size"

        szl = int(np.prod(self.shape[0:n]))
        szr = int(np.prod(self.shape[n + 1 :]))
        szn = self.shape[n]

        if n == 0:
            Ur = ttb.khatrirao(U[1 : self.ndims], reverse=True)
            Y = np.reshape(self.data, (szn, szr), order="F")
            return Y @ Ur
        if n == self.ndims - 1:  # pylint: disable=no-else-return
            Ul = ttb.khatrirao(U[0 : self.ndims - 1], reverse=True)
            Y = np.reshape(self.data, (szl, szn), order="F")
            return Y.T @ Ul
        else:
            Ul = ttb.khatrirao(U[n + 1 :], reverse=True)
            Ur = np.reshape(ttb.khatrirao(U[0:n], reverse=True), (szl, 1, R), order="F")
            Y = np.reshape(self.data, (-1, szr), order="F")
            Y = Y @ Ul
            Y = np.reshape(Y, (szl, szn, R), order="F")
            V = np.zeros((szn, R))
            for r in range(R):
                V[:, [r]] = Y[:, :, r].T @ Ur[:, :, r]
            return V

    @property
    def ndims(self) -> int:
        """
        Return the number of dimensions of a tensor

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> X.ndims
        2
        """
        if self.shape == (0,):
            return 0
        return len(self.shape)

    @property
    def nnz(self) -> int:
        """
        Number of non-zero elements in tensor

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> X.nnz
        4
        """
        return np.count_nonzero(self.data)

    def norm(self) -> np.floating:
        """
        Frobenius Norm of Tensor

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> X.norm()
        2.0
        """
        # default of np.linalg.norm is to vectorize the data and compute the vector
        # norm, which is equivalent to the Frobenius norm for multidimensional arrays.
        # However, the argument 'fro' only works for 1-D and 2-D arrays currently.
        return np.linalg.norm(self.data)

    def nvecs(self, n: int, r: int, flipsign: bool = True) -> np.ndarray:
        """
        Compute the leading mode-n eigenvectors for a tensor

        Parameters
        ----------
        n: Mode to unfold
        r: Number of eigenvectors to compute
        flipsign: Make each eigenvector's largest element positive

        Examples
        --------
        >>> tensor1 = ttb.tensor.from_data(np.array([[1, 2], [3, 4]]))
        >>> tensor1.nvecs(0,1)  # doctest: +ELLIPSIS
        array([[0.4045...],
               [0.9145...]])
        >>> tensor1.nvecs(0,2)  # doctest: +ELLIPSIS
        array([[ 0.4045...,  0.9145...],
               [ 0.9145..., -0.4045...]])
        """
        Xn = ttb.tenmat.from_tensor_type(self, rdims=np.array([n])).double()
        y = Xn @ Xn.T

        if r < y.shape[0] - 1:
            w, v = scipy.sparse.linalg.eigsh(y, r)
            v = v[:, (-np.abs(w)).argsort()]
            v = v[:, :r]
        else:
            warnings.warn(
                "Greater than or equal to tensor.shape[n] - 1 eigenvectors"
                " requires cast to dense to solve"
            )
            w, v = scipy.linalg.eigh(y)
            v = v[:, (-np.abs(w)).argsort()]
            v = v[:, :r]

        if flipsign:
            idx = np.argmax(np.abs(v), axis=0)
            for i in range(v.shape[1]):
                if v[idx[i], i] < 0:
                    v[:, i] *= -1
        return v

    def permute(self, order: np.ndarray) -> tensor:
        """
        Permute tensor dimensions.

        Parameters
        ----------
        order: New order of tensor dimensions

        Returns
        -------
        Updated tensor with shapeNew == shapePrevious[order]

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> Y = X.permute(np.array((1,0)))
        >>> X.isequal(Y)
        True
        """
        if self.ndims != order.size:
            assert False, "Invalid permutation order"

        # If order is empty, return
        if order.size == 0:
            return ttb.tensor.from_tensor_type(self)

        # Check for special case of an order-1 object, has no effect
        if (order == 1).all():
            return ttb.tensor.from_tensor_type(self)

        # Np transpose does error checking on order, acts as permutation
        return ttb.tensor.from_data(np.transpose(self.data, order))

    def reshape(self, shape: Tuple[int, ...]) -> tensor:
        """
        Reshapes a tensor

        Parameters
        ----------
        shape: New shape

        Examples
        --------
        >>> X = ttb.tensor.from_data(np.ones((2,2)))
        >>> Y = X.reshape((4,1))
        >>> Y.shape
        (4, 1)
        """

        if np.prod(self.shape) != np.prod(shape):
            assert False, "Reshaping a tensor cannot change number of elements"

        return ttb.tensor.from_data(np.reshape(self.data, shape, order="F"), shape)

    def squeeze(self) -> Union[tensor, np.ndarray, float]:
        """
        Removes singleton dimensions from a tensor

        Returns
        -------
        Tensor or scalar if all dims squeezed

        Examples
        --------
        >>> tensor1 = ttb.tensor.from_data(np.array([[[4]]]))
        >>> tensor1.squeeze()
        4
        >>> tensor2 = ttb.tensor.from_data(np.array([[1, 2, 3]]))
        >>> tensor2.squeeze().data
        array([1, 2, 3])

        """
        shapeArray = np.array(self.shape)
        if np.all(shapeArray > 1):  # pylint: disable=no-else-return
            return ttb.tensor.from_tensor_type(self)
        else:
            idx = np.where(shapeArray > 1)
            if idx[0].size == 0:
                return np.squeeze(self.data)[()]
            return ttb.tensor.from_data(np.squeeze(self.data))

    def symmetrize(
        self, grps: Optional[np.ndarray] = None, version: Optional[Any] = None
    ) -> (
        tensor
    ):  # pylint: disable=too-many-branches, too-many-statements, too-many-locals
        """
        Symmetrize a tensor in the specified modes
        Notes
        -----
        It is *the same or less* work to just call X = symmetrize(X) then to first
        check if X is symmetric and then symmetrize it, even if X is already symmetric.

        Parameters
        ----------
        grps: Modes to check for symmetry
        version: Any non-None value will call the non-default old version
        Returns
        -------
        """
        n = self.ndims
        sz = np.array(self.shape)

        if grps is None:
            grps = np.arange(0, n)

        if len(grps.shape) == 1:
            grps = np.array([grps])

        data = self.data.copy()

        # Use default newer faster version
        if version is None:  # pylint: disable=no-else-return
            ngrps = len(grps)
            for i in range(0, ngrps):
                # Extract current group
                thisgrp = grps[i]

                # Check tensor dimensions first
                if not np.all(sz[thisgrp[0]] == sz[thisgrp]):
                    assert False, "Dimension mismatch for symmetrization"

                # Check for no overlap in the sets
                if i < ngrps - 1:
                    if not np.intersect1d(thisgrp, grps[i + 1 :, :]).size == 0:
                        assert False, "Cannot have overlapping symmetries"

                # Construct matrix ind where each row is the multi-index for one
                # element of tensor
                idx = ttb.tt_ind2sub(self.shape, np.arange(0, data.size))

                # Find reference index for every element in the tensor - this
                # is to its index in the symmetrized tensor. This puts every
                # element into a 'class' of entries that will be the same under
                # symmetry.

                classidx = idx
                classidx[:, thisgrp] = np.sort(idx[:, thisgrp], axis=1)
                linclassidx = ttb.tt_sub2ind(self.shape, classidx)

                # Compare each element to its class exemplar
                if np.all(data.ravel() == data[tuple(classidx.transpose())]):
                    continue

                # Take average over all elements in the same class
                classSum = accumarray(linclassidx, data.ravel())
                classNum = accumarray(linclassidx, 1)
                # We ignore this division error state because if we don't have an entry
                # in linclassidx we won't reference the inf or nan in the slice below
                with np.errstate(divide="ignore", invalid="ignore"):
                    avg = classSum / classNum

                newdata = avg[linclassidx]
                data = np.reshape(newdata, self.shape)

            return ttb.tensor.from_data(data)

        else:  # Original version
            # Check tensor dimensions for compatibility with symmetrization
            ngrps = len(grps)
            for i in range(0, ngrps):
                dims = grps[i]
                for j in dims[1:]:
                    if sz[j] != sz[dims[0]]:
                        assert False, "Dimension mismatch for symmetrization"

            # Check for no overlap in sets
            for i in range(0, ngrps):
                for j in range(i + 1, ngrps):
                    if not np.intersect1d(grps[i, :], grps[j, :]).size == 0:
                        assert False, "Cannot have overlapping symmetries"

            # Create the combinations for each symmetrized subset
            combos = []
            for i in range(0, ngrps):
                combos.append(np.array(list(permutations(grps[i, :]))))
            combos = np.stack(combos)

            # Create all the permuations to be averaged
            combo_lengths = [len(perm) for perm in combos]
            total_perms = int(np.prod(combo_lengths))
            sym_perms = np.tile(np.arange(0, n), [total_perms, 1])
            for i in range(0, ngrps):
                ntimes = np.prod(combo_lengths[0:i], dtype=int)
                ncopies = np.prod(combo_lengths[i + 1 :], dtype=int)
                nelems = len(combos[i])

                perm_idx = 0
                for j in range(0, ntimes):
                    for k in range(0, nelems):
                        for _ in range(0, ncopies):
                            # TODO: Does this do anything? Matches MATLAB
                            # at very least should be able to flatten
                            sym_perms[perm_idx, grps[i]] = combos[i][k, :]
                            perm_idx += 1

            # Create an average tensor
            Y = ttb.tensor.from_data(np.zeros(self.shape))
            for i in range(0, total_perms):
                Y += self.permute(sym_perms[i, :])

            Y /= total_perms

            # It's not *exactly* symmetric due to oddities in differently ordered
            # summations and so on, so let's fix that.
            # Idea borrowed from Gergana Bounova:
            # http://www.mit.edu/~gerganaa/downloads/matlab/symmetrize.m
            for i in range(0, total_perms):
                Z = Y.permute(sym_perms[i, :])
                Y.data[:] = np.maximum(Y.data[:], Z.data[:])

            return Y

    def ttm(
        self,
        matrix: Union[np.ndarray, List[np.ndarray]],
        dims: Optional[Union[float, np.ndarray]] = None,
        exclude_dims: Optional[Union[int, np.ndarray]] = None,
        transpose: bool = False,
    ) -> tensor:
        """
        Tensor times matrix

        Parameters
        ----------
        matrix: Matrix or matrices to multiple by
        dims: Dimensions to multiply against
        exclude_dims: Use all dimensions but these
        transpose: Transpose matrices during multiplication
        """
        if dims is None and exclude_dims is None:
            dims = np.arange(self.ndims)
        elif isinstance(dims, list):
            dims = np.array(dims)
        elif isinstance(dims, (float, int, np.generic)):
            dims = np.array([dims])

        if isinstance(exclude_dims, (float, int)):
            exclude_dims = np.array([exclude_dims])

        if isinstance(matrix, list):
            # Check that the dimensions are valid
            dims, vidx = ttb.tt_dimscheck(self.ndims, len(matrix), dims, exclude_dims)

            # Calculate individual products
            Y = self.ttm(matrix[vidx[0]], dims[0], transpose=transpose)
            for k in range(1, dims.size):
                Y = Y.ttm(matrix[vidx[k]], dims[k], transpose=transpose)
            return Y

        if not isinstance(matrix, np.ndarray):
            assert False, f"matrix must be of type numpy.ndarray but got:\n{matrix}"

        dims, _ = ttb.tt_dimscheck(self.ndims, dims=dims, exclude_dims=exclude_dims)

        if not (dims.size == 1 and np.isin(dims, np.arange(self.ndims))):
            assert False, "dims must contain values in [0,self.dims)"

        # old version (ver=0)
        shape = np.array(self.shape, dtype=int)
        n = dims[0]
        order = np.array([n] + list(range(0, n)) + list(range(n + 1, self.ndims)))
        newdata = self.permute(order).data
        ids = np.array(list(range(0, n)) + list(range(n + 1, self.ndims)))
        second_dim = 1
        if len(ids) > 0:
            second_dim = np.prod(shape[ids])
        newdata = np.reshape(newdata, (shape[n], second_dim), order="F")
        if transpose:
            newdata = matrix.T @ newdata
            p = matrix.shape[1]
        else:
            newdata = matrix @ newdata
            p = matrix.shape[0]

        newshape = np.array(
            [p] + list(shape[range(0, n)]) + list(shape[range(n + 1, self.ndims)])
        )
        Y_data = np.reshape(newdata, newshape, order="F")
        Y_data = np.transpose(Y_data, np.argsort(order))
        return ttb.tensor.from_data(Y_data)

    def ttt(
        self,
        other: tensor,
        selfdims: Optional[Union[int, np.ndarray]] = None,
        otherdims: Optional[Union[int, np.ndarray]] = None,
    ) -> tensor:
        """
        Tensor multiplication (tensor times tensor)

        Parameters
        ----------
        other: Tensor to multiply by
        selfdims: Dimensions to contract this tensor by for multiplication
        otherdims: Dimensions to contract other tensor by for multiplication
        """

        if not isinstance(other, tensor):
            assert False, "other must be of type tensor"

        if selfdims is None:
            selfdims = np.array([], dtype=int)
        elif isinstance(selfdims, int):
            selfdims = np.array([selfdims])
        selfshape = tuple(np.array(self.shape)[selfdims])

        if otherdims is None:
            otherdims = selfdims.copy()
        elif isinstance(otherdims, int):
            otherdims = np.array([otherdims])
        othershape = tuple(np.array(other.shape)[otherdims])

        if np.any(selfshape != othershape):
            assert (
                False
            ), f"Specified dimensions do not match got {selfshape} and {othershape}"

        # Compute the product

        # Avoid transpose by reshaping self and computing result = self * other
        amatrix = ttb.tenmat.from_tensor_type(self, cdims=selfdims)
        bmatrix = ttb.tenmat.from_tensor_type(other, rdims=otherdims)
        cmatrix = amatrix * bmatrix

        # Check whether or not the result is a scalar
        if isinstance(cmatrix, ttb.tenmat):
            return ttb.tensor.from_tensor_type(cmatrix)
        return cmatrix

    def ttv(
        self,
        vector: Union[np.ndarray, List[np.ndarray]],
        dims: Optional[Union[int, np.ndarray]] = None,
        exclude_dims: Optional[Union[int, np.ndarray]] = None,
    ) -> tensor:
        """
        Tensor times vector

        Parameters
        ----------
        vector: Vector(s) to multiply against
        dims: Dimensions to multiply with vector(s)
        exclude_dims: Use all dimensions but these
        """

        if dims is None and exclude_dims is None:
            dims = np.array([])
        elif isinstance(dims, (float, int)):
            dims = np.array([dims])

        if isinstance(exclude_dims, (float, int)):
            exclude_dims = np.array([exclude_dims])

        # Check that vector is a list of vectors, if not place single vector as element
        # in list
        if len(vector) > 0 and isinstance(vector[0], (int, float, np.int_, np.float_)):
            return self.ttv(np.array([vector]), dims, exclude_dims)

        # Get sorted dims and index for multiplicands
        dims, vidx = ttb.tt_dimscheck(self.ndims, len(vector), dims, exclude_dims)

        # Check that each multiplicand is the right size.
        for i in range(dims.size):
            if vector[vidx[i]].shape != (self.shape[dims[i]],):
                assert False, "Multiplicand is wrong size"

        # Extract the data
        c = self.data.copy()

        # Permute it so that the dimensions we're working with come last
        remdims = np.setdiff1d(np.arange(0, self.ndims), dims)
        if self.ndims > 1:
            c = np.transpose(c, np.concatenate((remdims, dims)))

        # Do each multiply in sequence, doing the highest index first, which is
        # important for vector multiplies.
        n = self.ndims
        sz = np.array(self.shape)[np.concatenate((remdims, dims))]

        for i in range(dims.size - 1, -1, -1):
            c = np.reshape(c, tuple([np.prod(sz[0 : n - 1]), sz[n - 1]]), order="F")
            c = c.dot(vector[vidx[i]])
            n -= 1
        # If needed, convert the final result back to tensor
        if n > 0:
            return ttb.tensor.from_data(c, tuple(sz[0:n]))
        return c[0]

    def ttsv(
        self,
        vector: Union[np.ndarray, List[np.ndarray]],
        skip_dim: Optional[int] = None,
        version: Optional[int] = None,
    ) -> Union[np.ndarray, tensor]:
        """
        Tensor times same vector in multiple modes

        Parameters
        ----------
        vector: Vector(s) to multiply against
        skip_dim: Multiply tensor by vector in all dims except [0, skip_dim]
        """
        # Only two simple cases are supported
        if skip_dim is None:
            exclude_dims = None
            skip_dim = -1  # For easier math later
        elif skip_dim < 0:
            raise ValueError("Invalid modes in ttsv")
        else:
            exclude_dims = np.arange(0, skip_dim + 1)

        if version == 1:  # Calculate the old way
            P = self.ndims
            X = np.array([vector for i in range(P)])
            if skip_dim in (0, 1):  # Return scalar or matrix
                return self.ttv(X, exclude_dims=exclude_dims).double()
            return self.ttv(X, exclude_dims=exclude_dims)

        if version == 2 or version is None:  # Calculate the new way
            d = self.ndims
            sz = self.shape[0]  # Sizes of all modes must be the same

            # pylint: disable=invalid-unary-operand-type
            dnew = skip_dim + 1  # Number of modes in result
            drem = d - dnew  # Number of modes multiplied out

            y = self.data
            for i in range(drem, 0, -1):
                yy = np.reshape(y, (sz ** (dnew + i - 1), sz), order="F")
                y = yy.dot(vector)

            # Convert to matrix if 2-way or convert back to tensor if result is >= 3-way
            if dnew == 2:
                return np.reshape(y, [sz, sz], order="F")
            if dnew > 2:
                return ttb.tensor.from_data(
                    np.reshape(y, newshape=sz * np.ones(dnew, dtype=int), order="F")
                )
            return y
        assert False, "Invalid value for version; should be None, 1, or 2"

    def __setitem__(
        self, key, value
    ):  # pylint: disable=too-many-branches, too-many-statements
        """
        SUBSASGN Subscripted assignment for a tensor.

        We can assign elements to a tensor in three ways.

        Case 1: X(R1,R2,...,RN) = Y, in which case we replace the
        rectangular subtensor (or single element) specified by the ranges
        R1,...,RN with Y. The right-hand-side can be a scalar, a tensor, or an
        MDA.

        Case 2a: X(S) = V, where S is a p x n array of subscripts and V is
        a scalar or a vector containing p values.

        Case 2b: X(I) = V, where I is a set of p linear indices and V is a
        scalar or a vector containing p values. Resize is not allowed in this
        case.

        Examples
        X = tensor(rand(3,4,2))
        X(1:2,1:2,1) = ones(2,2) <-- replaces subtensor
        X([1 1 1;1 1 2]) = [5;7] <-- replaces two elements
        X([1;13]) = [5;7] <-- does the same thing
        X(1,1,2:3) = 1 <-- grows tensor
        X(1,1,4) = 1 %<- grows the size of the tensor
        """
        # Figure out if we are doing a subtensor, a list of subscripts or a list of
        # linear indices
        access_type = "error"
        if self.ndims <= 1:
            if isinstance(key, np.ndarray):
                access_type = "subscripts"
            else:
                access_type = "subtensor"
        else:
            if isinstance(key, np.ndarray):
                if len(key.shape) > 1 and key.shape[1] >= self.ndims:
                    access_type = "subscripts"
                elif len(key.shape) == 1 or key.shape[1] == 1:
                    access_type = "linear indices"
            elif isinstance(key, tuple):
                validSubtensor = [
                    isinstance(keyElement, (int, slice)) for keyElement in key
                ]
                if np.all(validSubtensor):
                    access_type = "subtensor"

        # Case 1: Rectangular Subtensor
        if access_type == "subtensor":
            # Extract array of subscripts
            subs = key

            # Will the size change? If so we first need to resize x
            n = self.ndims
            sliceCheck = []
            for element in subs:
                if isinstance(element, slice):
                    if element.stop is None:
                        sliceCheck.append(1)
                    else:
                        sliceCheck.append(element.stop)
                else:
                    sliceCheck.append(element)
            bsiz = np.array(sliceCheck)
            if n == 0:
                newsiz = (bsiz[n:] + 1).astype(int)
            else:
                newsiz = np.concatenate(
                    (np.max((self.shape, bsiz[0:n] + 1), axis=0), bsiz[n:] + 1)
                ).astype(int)
            if (newsiz != self.shape).any():
                # We need to enlarge x.data.
                newData = np.zeros(shape=tuple(newsiz))
                idx = [slice(None, currentShape) for currentShape in self.shape]
                if self.data.size > 0:
                    newData[tuple(idx)] = self.data
                self.data = newData

                self.shape = tuple(newsiz)
            if isinstance(value, ttb.tensor):
                self.data[key] = value.data
            else:
                self.data[key] = value

            return

        # Case 2a: Subscript indexing
        if access_type == "subscripts":
            # Extract array of subscripts
            subs = key

            # Will the size change? If so we first need to resize x
            n = self.ndims
            if (
                len(subs.shape) == 1
                and len(self.shape) == 1
                and self.shape[0] < subs.shape[0]
            ):
                bsiz = subs
            elif len(subs.shape) == 1:
                bsiz = np.array([np.max(subs, axis=0)])
                key = key.tolist()
            else:
                bsiz = np.array(np.max(subs, axis=0))
            if n == 0:
                newsiz = (bsiz[n:] + 1).astype(int)
            else:
                newsiz = np.concatenate(
                    (np.max((self.shape, bsiz[0:n] + 1), axis=0), bsiz[n:] + 1)
                ).astype(int)

            if (newsiz != self.shape).any():
                # We need to enlarge x.data.
                newData = np.zeros(shape=tuple(newsiz))
                idx = [slice(None, currentShape) for currentShape in self.shape]
                if self.data.size > 0:
                    newData[idx] = self.data
                self.data = newData

                self.shape = tuple(newsiz)

            # Finally we can copy in new data
            if isinstance(key, list):
                self.data[key] = value
            elif key.shape[0] == 1:  # and len(key.shape) == 1:
                self.data[tuple(key[0, :])] = value
            else:
                self.data[tuple(key)] = value
            return

        # Case 2b: Linear Indexing
        if access_type == "linear indices":
            idx = key
            if (idx > np.prod(self.shape)).any():
                assert (
                    False
                ), "TTB:BadIndex In assignment X[I] = Y, a tensor X cannot be resized"
            idx = tt_ind2sub(self.shape, idx)
            if idx.shape[0] == 1:
                self.data[tuple(idx[0, :])] = value
            else:
                actualIdx = tuple(idx.transpose())
                self.data[actualIdx] = value
            return

        assert False, "Invalid use of tensor setitem"

    def __getitem__(self, item):
        """
        SUBSREF Subscripted reference for tensors.

        We can extract elements or subtensors from a tensor in the
        following ways.

        Case 1a: y = X(i1,i2,...,iN), where each in is an index, returns a
        scalar.

        Case 1b: Y = X(R1,R2,...,RN), where one or more Rn is a range and
        the rest are indices, returns a sparse tensor.

        Case 2a: V = X(S) or V = X(S,'extract'), where S is a p x n array
        of subscripts, returns a vector of p values.

        Case 2b: V = X(I) or V = X(I,'extract'), where I is a set of p
        linear indices, returns a vector of p values.

        Any ambiguity results in executing the first valid case. This
        is particularly an issue if ndims(X)==1.

        Examples
        X = tensor(rand(3,4,2,1),[3 4 2 1]);
        X.data <-- returns multidimensional array
        X.size <-- returns size
        X(1,1,1,1) <-- produces a scalar
        X(1,1,1,:) <-- produces a tensor of order 1 and size 1
        X(:,1,1,:) <-- produces a tensor of size 3 x 1
        X(1:2,[2 4],1,:) <-- produces a tensor of size 2 x 2 x 1
        X(1:2,[2 4],1,1) <-- produces a tensor of size 2 x 2
        X([1,1,1,1;3,4,2,1]) <-- returns a vector of length 2
        X = tensor(rand(10,1),10);
        X([1:6]') <-- extracts a subtensor
        X([1:6]','extract') <-- extracts a vector of 6 elements

        Parameters
        ----------

        Returns
        -------
        :class:`pyttb.tensor` or :class:`numpy.ndarray`
        """
        # Case 1: Rectangular Subtensor
        if (
            isinstance(item, tuple)
            and len(item) == self.ndims
            and item[len(item) - 1] != "extract"
        ):
            # Copy the subscripts
            region = item

            # Extract the data
            newdata = self.data[region]

            # Determine the subscripts
            newsiz = []  # future new size
            kpdims = []  # dimensions to keep
            rmdims = []  # dimensions to remove

            # Determine the new size and what dimensions to keep
            for i, a_region in enumerate(region):
                if isinstance(a_region, slice):
                    newsiz.append(self.shape[i])
                    kpdims.append(i)
                elif not isinstance(a_region, int) and len(a_region) > 1:
                    newsiz.append(np.prod(a_region))
                    kpdims.append(i)
                else:
                    rmdims.append(i)

            newsiz = np.array(newsiz, dtype=int)
            kpdims = np.array(kpdims, dtype=int)
            rmdims = np.array(rmdims, dtype=int)

            # If the size is zero, then the result is returned as a scalar
            # otherwise, we convert the result to a tensor
            if newsiz.size == 0:
                a = newdata
            else:
                a = ttb.tensor.from_data(newdata)
            return a

        # *** CASE 2a: Subscript indexing ***
        if len(item) > 1 and isinstance(item[-1], str) and item[-1] == "extract":
            # Extract array of subscripts
            subs = np.array(item[0])
            a = np.squeeze(self.data[tuple(subs.transpose())])
            # TODO if is row make column?
            return ttb.tt_subsubsref(a, subs)

        # Case 2b: Linear Indexing
        if len(item) >= 2 and not isinstance(item[-1], str):
            assert False, "Linear indexing requires single input array"
        idx = item[0]
        a = np.squeeze(self.data[tuple(ttb.tt_ind2sub(self.shape, idx).transpose())])
        # Todo if row make column?
        return ttb.tt_subsubsref(a, idx)

    def __eq__(self, other):
        """
        Equal for tensors

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def tensor_equality(x, y):
            return x == y

        return ttb.tt_tenfun(tensor_equality, self, other)

    def __ne__(self, other):
        """
        Not equal (!=) for tensors

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def tensor_not_equal(x, y):
            return x != y

        return ttb.tt_tenfun(tensor_not_equal, self, other)

    def __ge__(self, other):
        """
        Greater than or equal (>=) for tensors

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def greater_or_equal(x, y):
            return x >= y

        return ttb.tt_tenfun(greater_or_equal, self, other)

    def __le__(self, other):
        """
        Less than or equal (<=) for tensors

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def less_or_equal(x, y):
            return x <= y

        return ttb.tt_tenfun(less_or_equal, self, other)

    def __gt__(self, other):
        """
        Greater than (>) for tensors

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def greater(x, y):
            return x > y

        return ttb.tt_tenfun(greater, self, other)

    def __lt__(self, other):
        """
        Less than (<) for tensors

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def less(x, y):
            return x < y

        return ttb.tt_tenfun(less, self, other)

    def __sub__(self, other):
        """
        Binary subtraction (-) for tensors

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def minus(x, y):
            return x - y

        return ttb.tt_tenfun(minus, self, other)

    def __add__(self, other):
        """
        Binary addition (+) for tensors

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        # If rhs is sumtensor, treat as such
        if isinstance(other, ttb.sumtensor):  # pragma: no cover
            return other.__add__(self)

        def tensor_add(x, y):
            return x + y

        return ttb.tt_tenfun(tensor_add, self, other)

    def __radd__(self, other):
        """
        Right binary addition (+) for tensors

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        return self.__add__(other)

    def __pow__(self, power):
        """
        Element Wise Power (**) for tensors

        Parameters
        ----------
        other::class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def tensor_pow(x, y):
            return x**y

        return ttb.tt_tenfun(tensor_pow, self, power)

    def __mul__(self, other):
        """
        Element wise multiplication (*) for tensors, self*other

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def mul(x, y):
            return x * y

        if isinstance(other, (ttb.ktensor, ttb.sptensor, ttb.ttensor)):
            other = other.full()

        return ttb.tt_tenfun(mul, self, other)

    def __rmul__(self, other):
        """
        Element wise right multiplication (*) for tensors, other*self

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """
        return self.__mul__(other)

    def __truediv__(self, other):
        """
        Element wise left division (/) for tensors, self/other

        Parameters
        ----------
        other: :class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def div(x, y):
            # We ignore the divide by zero errors because np.inf/np.nan is an
            # appropriate representation
            with np.errstate(divide="ignore", invalid="ignore"):
                return x / y

        return ttb.tt_tenfun(div, self, other)

    def __rtruediv__(self, other):
        """
        Element wise right division (/) for tensors, other/self

        Parameters
        ----------
        other::class:`pyttb.tensor`, float, int

        Returns
        -------
        :class:`pyttb.tensor`
        """

        def div(x, y):
            # We ignore the divide by zero errors because np.inf/np.nan is an
            # appropriate representation
            with np.errstate(divide="ignore", invalid="ignore"):
                return x / y

        return ttb.tt_tenfun(div, other, self)

    def __pos__(self):
        """
        Unary plus (+) for tensors

        Returns
        -------
        :class:`pyttb.tensor`
            copy of tensor
        """

        return ttb.tensor.from_data(self.data)

    def __neg__(self):
        """
        Unary minus (-) for tensors

        Returns
        -------
        :class:`pyttb.tensor`
            copy of tensor
        """

        return ttb.tensor.from_data(-1 * self.data)

    def __repr__(self):
        """
        String representation of a tensor.

        Returns
        -------
        str
            Contains the shape and data as strings on different lines.
        """
        if self.ndims == 0:
            s = ""
            s += "empty tensor of shape "
            s += str(self.shape)
            s += "\n"
            s += "data = []"
            return s

        s = ""
        s += "tensor of shape "
        s += (" x ").join([str(int(d)) for d in self.shape])
        s += "\n"

        if self.ndims == 1:
            s += "data"
            if self.ndims == 1:
                s += "[:]"
                s += " = \n"
                s += str(self.data)
                s += "\n"
                return s
        for i in np.arange(np.prod(self.shape[:-2])):
            s += "data"
            if self.ndims == 2:
                s += "[:, :]"
                s += " = \n"
                s += str(self.data)
                s += "\n"
            elif self.ndims > 2:
                idx = ttb.tt_ind2sub(self.shape[:-2], np.array([i]))
                s += str(idx[0].tolist())[0:-1]
                s += ", :, :]"
                s += " = \n"
                s += str(
                    self.data[
                        tuple(
                            np.concatenate(
                                (idx[0], np.array([slice(None), slice(None)]))
                            )
                        )
                    ]
                )
                s += "\n"
        # s += '\n'
        return s

    __str__ = __repr__


if __name__ == "__main__":
    import doctest  # pragma: no cover

    doctest.testmod()  # pragma: no cover
